#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes.utils import utils
import shutil, math, time

# the system control class.
class Metrics(object):
	def __init__(self, database=None):

		# docs.
		DOCS = {
			"module":"website.metrics", 
			"initialized":True,
			"description":[], 
			"chapter": "Website", }

		# history.
		self.history = 7 # in days.

		# attributes.
		self.database = database
		if isinstance(self.database, (str,String,FilePath)):
			self.database = Directory(str(self.database))

		# files.
		if not Files.exists(self.database.join("data/metrics")): Files.create(self.database.join("data/metrics"), directory=True)
		self.__requests__ = Dictionary(path=self.database.join("data/metrics/requests", ""), load=True, default={})
		self.__saving_requests__ = False

		# clean.
		self.clean()

		#
	def clean(self):


		# cleanings.
		date = Date()
		decreased = date.decrease(date.timestamp, days=self.history)
		for id, requests in self.__requests__.items():
			for timestamp in list(requests.keys()):
				if date.compare(timestamp, decreased) in ["past"]:
					del self.__requests__[id][str(timestamp)]

		# save cleaned.
		self.__requests__.save()

		# handler.
		return dev0s.response.success("Successfully cleaned the metrics data.")

		#

	# metrics functions.
	def requests(self):
		
		# get historical data.
		self.__safe_load__(self.__requests__)
		items = []
		for request_id in ["web", "api", "auth"]:
			try: self.__requests__[request_id]
			except KeyError: self.__requests__[request_id] = {}
			for key, value in self.__requests__[request_id].items():
				if key not in items: items.append(key)
		d = {}
		for timestamp in items:
			l = {}
			for request_id in ["web", "api", "auth"]:
				try: l[request_id] = self.__requests__[request_id][timestamp]
				except KeyError: l[request_id] = {}
			d[timestamp] = {
				"Web Requests":len(l["web"]),
				"API Requests":len(l["api"]),
				"Auth Requests":len(l["auth"]),
			}
		
		# graph data.
		graph = self.create_line_graph(
			data=d, 
			keep=["Web Requests", "API Requests", "Auth Requests"], 
			colors={
				"Web Requests":"#FD304E",
				"API Requests":"#323B83",
				"Auth Requests":"#006633",
			},
		)

		# handler.
		return dev0s.response.success("Successfully retrieved the requests data.", {
			"data":d,
			"graph":graph,
		})
	def auth_requests(self):
		
		# get historical data.
		self.__safe_load__(self.__requests__)
		d = {}
		try: self.__requests__["auth"]
		except KeyError: self.__requests__["auth"] = {}
		for timestamp, data in self.__requests__["auth"].items():
			success, failed = 0, 0 
			for item in data:
				if item["response"]["success"]: success += 1
				else: failed += 1
			d[timestamp] = {
				"Successfull":success,
				"Failed":failed,
			}
		
		# graph data.
		graph = self.create_line_graph(
			data=d, 
			keep=["Successfull", "Failed", "Auth Requests"], 
			colors={
				"Successfull":"#FD304E",
				"Failed":"#323B83",
			},
		)

		# handler.
		return dev0s.response.success("Successfully retrieved the auth requests data.", {
			"data":d,
			"graph":graph,
		})
	def disk_space(self, mode="GB"):
		def auto_size(size_bytes):
			if size_bytes == 0:
				return "0B"
			size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
			i = int(math.floor(math.log(size_bytes, 1024)))
			p = math.pow(1024, i)
			s = round(size_bytes / p, 2)
			return "%s %s" % (s, size_name[i])
		def size(size_bytes):
			if mode.upper() == "TB":
				return int(('{:,}'.format(int(round(size_bytes/1024**4,2))).replace(',', '.')).replace(".",""))
			elif mode.upper() == "GB":
				return int(('{:,}'.format(int(round(size_bytes/1024**3,2))).replace(',', '.')).replace(".",""))
			elif mode.upper() == "MB":
				return int(('{:,}'.format(int(round(size_bytes/1024**2,2))).replace(',', '.')).replace(".",""))
			elif mode.upper() == "KB":
				return int(('{:,}'.format(int(round(size_bytes/1024,2))).replace(',', '.')).replace(".",""))
			else:
				return int(('{:,}'.format(int(int(size_bytes))).replace(',', '.')).replace(".",""))
		total, used, free = shutil.disk_usage(str(self.database))
		return dev0s.response.success("Successfully retrieved the free disk space.", {
			"total":size(total), 
			"used":size(used), 
			"free":size(free),
			"graph":self.create_pie_graph(
				data={
					"Used (GB)":size(used),
					"Free (GB)":size(free),
				},
				colors={
					"Used (GB)": "#FD304E",
					"Free (GB)": "#323B8390",
				},
			),
		})
	
	# create graph function.
	def create_pie_graph(self, 
		# the data.
		data={
			"Used (GB)":15,
			"Free (GB)":3,
		},
		# the key's colors.
		colors={
			"Used (GB)": "#FD304E",
			"Free (GB)": "#323B8390",
		},
		# the keys to keep.
		keep=["*"],
		# fill background color.
		fill=False,
	):
		labels, new_data, background, border = [], [], [], []
		for key,value in data.items():
			if "*" in keep or key in keep:
				labels.append(key)
				new_data.append(value)
				background.append(colors[key])
				border.append(colors[key]+"90")
		return {
			"labels": labels,
			"datasets": [
				{
			        "data": new_data,
			        "backgroundColor":background,
			        "borderColor":border,
			        "fill": fill,
			    }
			],
		}
	def create_line_graph(self, 
		# the data.
		data={
			"$timestamp":{
				"active":10,
				"non_active":3,
			},
		}, 
		# the key's colors.
		colors={
			"active":"#323B83",
			"non_active":"#FD304E",
		},
		# the keys to keep.
		keep=["*"], 
		# fill background color.
		fill=False,
	):
		summed = {}
		for key, value in data.items():
			for _key_, _value_ in value.items():
				if "*" in keep or _key_ in keep:
					try:summed[_key_]
					except KeyError: summed[_key_] = []
					summed[_key_].append(_value_)
		datasets = []
		for key, value in summed.items():
			datasets.append({
				"label": key[0].upper() + key[1:],
				"backgroundColor": colors[key],
				"borderColor": colors[key]+"90",
				"fill": fill,
				"data": value
			})
		return {
			"labels": list(data.keys()),
			"datasets": datasets,
		}

	# count requests.
	def count_api_request(self, request, data={}):
		return self.count_request(request, id="api", save=random.randrange(0, 101) <= 10, data=data)
	def count_web_request(self, request, data={}):
		return self.count_request(request, id="web", save=random.randrange(0, 101) <= 10, data=data)
	def count_auth_request(self, request, data={}):
		return self.count_request(request, id="auth", save=True, data=data)
	def count_request(self, request, id=None, data={"url":None}, save=False):
		if id == None: raise dev0s.exceptions.InvalidUsage("Define parameter: id.")
		try: username = request.user.username
		except: username = None
		date = Date()
		timestamp = f"{date.timestamp[:-2]}{int(int(date.minute)/10)}0"
		self.__safe_load__(self.__requests__)
		try:self.__requests__[str(id)]
		except KeyError: self.__requests__[str(id)] = {}
		try:self.__requests__[str(id)][timestamp]
		except KeyError: self.__requests__[str(id)][timestamp] = []
		l_data = {
			"username":username,
			"ip":utils.get_client_ip(request), }
		for key,value in data.items(): l_data[key] = value
		self.__requests__[id][timestamp].append(l_data)
		if save and not self.__saving_requests__:
			self.__saving_requests__ = True
			self.__requests__.save()
			self.__saving_requests__ = False
		return dev0s.response.success(f"Successfully counted request [{id}].")

	# system functions.
	def __safe_load__(self, dict_object, reattempt=True):
		try:
			dict_object.load()
		except: 
			dict_object.dictionary = {}
			dict_object.save() 
	
	
#


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes import utils, views
from w3bsite.classes import defaults as _defaults_

# the authentication requests.
class Requests(_defaults_.Defaults):
	def __init__(self, 
		# passed Website.x objects.
		defaults=None,
	):
		# defaults.
		_defaults_.Defaults.__init__(self)
		self.assign(defaults.dict())

		# urlpatterns.
		if dev0s.system.env.get("MIGRATIONS", format=bool, default=False):
			self.urls = []
		else:
			send_code = self.SendCode(defaults=defaults)
			self.urls = views.build_urls([
				self.SignUp(defaults=defaults, send_code=send_code),
				self.SignIn(defaults=defaults),
				send_code,
				self.ResetPassword(defaults=defaults),
				self.Activate(defaults=defaults),
				self.Authenticated(defaults=defaults),
			])

	# sign up.
	class SignUp(views.Request):
		def __init__(self, defaults=None, send_code=None):
			views.Request.__init__(self, "requests/authentication/", "signup", website=defaults.website)
			self.assign(defaults.dict())
			self.send_code = send_code
		def request(self, request):

			# check overall rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="daily", limit=1000, reset_minutes=3600*24, increment=True)
			if not response.success: return self.response(response)

			# check signup rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="signup", limit=3, reset_minutes=3600*24, increment=False)
			if not response.success: return self.response(response)

			# retrieve params.
			parameters, response = self.parameters.get(request, [
				"username",
				"email",
				"password",
				"verify_password",])
			if not response.success: return self.response(response)
			optional_parameters, _ = self.parameters.get(request, {
				"name":None,
			})

			# make request.
			response = self.users.create(
				email=parameters["email"],
				username=parameters["username"],
				password=parameters["password"],
				verify_password=parameters["verify_password"],
				name=optional_parameters["name"],)
			if not response.success: return self.response(response)

			# increment.
			_response_ = self.rate_limit.increment(ip=utils.get_client_ip(request), mode="signup")
			if not _response_.success: return _response_


			# send activation code.
			_response_ = self.send_code.send_code(username=parameters["username"], mode="activation", request=request)
			if not _response_.success: return self.response(_response_)

			# return create response.
			if response.success:
				del response["user"]
			return self.response(response)

			
			#

	# sign in.
	class SignIn(views.Request):
		def __init__(self, defaults=None,):
			views.Request.__init__(self, "requests/authentication/", "signin", website=defaults.website)
			self.assign(defaults.dict())
		def request(self, request):

			# check overall rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="daily", limit=1000, reset_minutes=3600*24, increment=True)
			if not response.success: 
				self.website.metrics.count_auth_request(request, data={
					"url":self.url,
					"response":response,
				})
				return self.response(response)

			# check signin rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="signin", limit=10, reset_minutes=5, increment=True)
			if not response.success: 
				self.website.metrics.count_auth_request(request, data={
					"url":self.url,
					"response":response,
				})
				return self.response(response)

			# retrieve params.
			parameters, response = self.parameters.get(request, [
				"username",
				"password",])
			if not response.success: 
				self.website.metrics.count_auth_request(request, data={
					"url":self.url,
					"response":response,
				})
				return self.response(response)
			optional_parameters, _ = self.parameters.get(request, {
				"code":None,
			})

			# html
			html = ""
			if self._2fa:
				title = "Sign In - Verification Code"
				html = Files.load(f"{SOURCE_PATH}/classes/apps/authentication/mail/authentication.html")

			# make request.
			response = self.users.authenticate(
				username=parameters["username"],
				password=parameters["password"],
				_2fa_code=optional_parameters["code"],
				_2fa=self._2fa,
				html=html,
				request=request,)
			self.website.metrics.count_auth_request(request, data={
				"url":self.url,
				"response":response,
			})
			return self.response(response)
			
			#

	# sign out.
	class SignOut(views.Request):
		def __init__(self, defaults=None,):
			views.Request.__init__(self, "requests/authentication/", "signout", website=defaults.website)
			self.assign(defaults.dict())
		def request(self, request):

			# make request.
			return self.response(self.users.signout(
				request=request,))
			
			#

	# send reset password email.
	class SendCode(views.Request):
		def __init__(self, defaults=None,):
			views.Request.__init__(self, "requests/authentication/", "send_code", website=defaults.website)
			self.assign(defaults.dict())
		def request(self, request):

			# check overall rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="daily", limit=1000, reset_minutes=3600*24, increment=True)
			if not response.success: return self.response(response)

			# check 2fa rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="2fa", limit=15, reset_minutes=3600, increment=True)
			if not response.success: return self.response(response)

			# retrieve params.
			parameters, response = self.parameters.get(request, [
				"username",
				"mode",])
			if not response.success: return self.response(response)
			if parameters["username"] in ["undefined", "null", "None", "none", ""]:
				return self.error("Refresh the web page.")

			# make request.
			return self.response(self.send_code(
				username=parameters["username"],
				mode=parameters['mode'],
				request=request,))
			
			#

			#
		def send_code(self, username=None, mode=None, request=None):

			# check library.
			#if not Files.exists("templates/mail"): os.mkdir("templates/mail")
			#response = utils.__check_package_files__([
			#	["templates/mail/reset_password.html","w3bsite","/classes/requests/mail/reset_password.html"],
			#	["templates/mail/authentication.html","w3bsite","/classes/requests/mail/authentication.html"],
			#	["templates/mail/activation.html","w3bsite","/classes/requests/mail/activation.html"],
			#])
			#if not response.success: return response

			# set mode.
			code = Integer(0).generate(length=6)
			if mode == "reset_password":
				title = "Reset Password - Verification Code"
				path = f"{SOURCE_PATH}/classes/apps/authentication/mail/reset_password.html"
			elif mode == "authentication":
				title = "Sign In - Verification Code"
				path = f"{SOURCE_PATH}/classes/apps/authentication/mail/authentication.html"
			elif mode == "activation":
				title = "Account Activation - Verification Code"
				path = f"{SOURCE_PATH}/classes/apps/authentication/mail/activation.html"
			else:
				return dev0s.response.error("Selected an invalid mode.")

			# parse html.
			ip = utils.get_client_ip(request)
			html = Files.load(path)

			# make request.
			return self.users.send_code(
				username=username,
				mode=mode,
				ip=ip,
				code=code,
				html=html,
				title=title,)

	# reset password.
	class ResetPassword(views.Request):
		def __init__(self, defaults=None,):
			views.Request.__init__(self, "requests/authentication/", "reset_password", website=defaults.website)
			self.assign(defaults.dict())
		def request(self, request):

			# check overall rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="daily", limit=1000, reset_minutes=3600*24, increment=True)
			if not response.success: return self.response(response)

			# retrieve params.
			parameters, response = self.parameters.get(request, [
				"code",
				"email",
				"password",
				"verify_password",])
			if not response.success: return self.response(response)
			if parameters["email"] in ["undefined", "null", "None", "none", ""]:
				return self.error("Refresh the web page.")

			# make request.
			response = self.users.verify_code(
				email=parameters["email"],
				code=parameters["code"],
				mode="reset_password",)
			if not response.success: return self.response(response)

			# make request.
			response = self.users.update(
				email=parameters["email"], # for id.
				password=parameters["password"],
				verify_password=parameters["verify_password"],)
			if response.success: 
				response["message"] = f"Succesfully resetted the password of user [{user.email}]."
				_response_ = self.users.save_password(email=user.email, username=user.username, password=parameters["password"])
				if not _response_.success: return self.response(_response_)
			return self.response(response)
			
			#

	# activate account.
	class Activate(views.Request):
		def __init__(self, defaults=None,):
			views.Request.__init__(self, "requests/authentication/", "activate", website=defaults.website)
			self.assign(defaults.dict())
		def request(self, request):

			# check overall rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="daily", limit=1000, reset_minutes=3600*24, increment=True)
			if not response.success: return self.response(response)

			# retrieve params.
			parameters, response = self.parameters.get(request, [
				"code",
				"email",])
			if not response.success: return self.response(response)
			if parameters["email"] in ["undefined", "null", "None", "none", ""]:
				return self.error("Refresh the web page.")

			# make request.
			response = self.users.verify_code(
				email=parameters["email"],
				code=parameters["code"],
				mode="activation",)
			if not response.success: return self.response(response)

			# make request.
			response = self.users.update(
				email=parameters["email"],
				email_verified=True,)
			if response.success: response["message"] = f"Succesfully activated the account user [{user.email}]."
			return self.response(response)
			
			#

	# is authenticated.
	class Authenticated(views.Request):
		def __init__(self, defaults=None,):
			views.Request.__init__(self, "requests/authentication/", "authenticated", website=defaults.website)
			self.assign(defaults.dict())
		def request(self, request):

			# check overall rate limit.
			response = self.rate_limit.verify(ip=utils.get_client_ip(request), mode="daily", limit=1000, reset_minutes=3600*24, increment=True)
			if not response.success: return self.response(response)

			# handler.
			try:
				email = request.user.email
			except AttributeError:
				email = None
			return self.success("Successfully checked if the user is authenticated.", {
				"authenticated":request.user.username != None and request.user.is_authenticated == True,
				"username":request.user.username,
				"email":email,
			})
			
			#

		#

	#

#


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
if not dev0s.env.get("MIGRATIONS", format=bool, default=False):
	from django.contrib.auth.models import User as DjangoUser
	from django.contrib.auth import authenticate, login
	from django.contrib.auth import login as _login_
	from django.contrib.auth import logout as _logout_
try:
	from w3bsite.classes.firebase import *
	from firebase_admin import credentials, auth, firestore, _auth_utils
except ModuleNotFoundError: a=1
from w3bsite.classes import defaults as _defaults_
from w3bsite.classes import email, utils

# the users class.
class Users(_defaults_.Defaults):
	def __init__(self,
		# noreply email settings.
		email_enabled=True,
		email_address=None, 
		email_password=None, 
		smtp_host="smtp.gmail.com", 
		smtp_port=587, 
		# objects.
		db=None,
		firestore=None, 
		stripe=None, 
		django=None,
		# defaults.
		defaults=None,
	):	

		# docs.
		DOCS = {
			"module":"website.users", 
			"initialized":True,
			"description":[
				"The <attribute>website.users</attribute> object class.",
				"Used for managing the website's users.",
			], 
			"chapter": "Users", }

		# defaults.
		_defaults_.Defaults.__init__(self, traceback="w3bsite.Website.users",)
		self.assign(defaults.dict())

		# variables.
		self.email_enabled = email_enabled
		self.email_address = email_address
		self.email_password = email_password
		self.smtp_host = smtp_host 
		self.smtp_port = smtp_port 
		self.visible_email = None

		# objects.
		self.db = db
		self.firestore = firestore
		self.stripe = stripe
		self.django = django

		# default user data.
		self.default_user_data = {
			"account": {
				"username":None,
				"email":None,
				"name":None,
				"password":None,
			},
			# the permissions.
			"permissions": {
				"activated":False,
				"email":True,
			},
			# timestamps.
			"timestamps": {
				"created":None,
			},
			# the keys.
			"keys": {
				"api_key":None,
			},
			# the requests.
			"requests": {
			},
		}

		# sys vars.
		self.email = None
		self.__timestamps__ = Dictionary({})
		self.__api_keys__ = Dictionary({})
		self.__subscriptions__ = Dictionary({})
		if self.live:
			self.__verification_codes__ = Dictionary(path=self.db.join("data/verification_codes"), default={}, load=self.live)
		else:
			self.__verification_codes__ = Dictionary(path=self.db.join("data/verification_codes"))

		#

	# functions.
	def get(self, 
		# define one of the following user id parameters.
		username=None,
		email=None,
	):	
		username, email = self.__correct_username_email__(username, email)
		return self.django.users.get(username=username, email=email)
	def create(self,
		# required:
		username=None,
		email=None,
		password=None,
		# optionals:
		verify_password=None,
		name=None,
		superuser=False,
		activated=True,
	):

		# check parameters.
		response = dev0s.response.parameters.check(
			traceback=self.__traceback__(function="create"),
			parameters={
				"username":username,
				"email":email,
				"password":password,
			})

		# check password.
		password = str(password)
		if len(password) < 8:
			return dev0s.response.error("The password must contain at least 8 characters.")
		elif password.lower() == password:
			return dev0s.response.error("The password must regular and capital letters.")
		elif verify_password != None and password != str(verify_password):
			return dev0s.response.error("Passwords do not match.")

		# create django user.
		if not self.exists(username=username, filter="django"):
			response = self.django.users.create(
				email=email,
				username=username,
				password=password,
				name=name,
				superuser=superuser,
				activate=activated,)
			if not response.success: return response
			user = response.user
		else:
			response = self.django.users.get(
				email=email,
				username=username,)
			if not response.success: return response
			user = response.user

		# try load existing data.
		print("Loading data...")
		response = self.load_data(email=email, username=username, create=True)
		if response.success and response.data != None:
			data = Dictionary(response.data).check(default=dict(self.default_user_data))
		else:
			data = dict(self.default_user_data)

		# save new user data.
		data["timestamps"]["created"] = Date().date
		data["keys"]["api_key"] = String().generate(length=48, capitalize=True, digits=True)
		_response_ = self.aes.encrypt(password)
		if not _response_.success: return _response_
		try:data["account"]
		except: data["account"] = {}
		data["account"]["username"] = username
		data["account"]["email"] = email
		data["account"]["name"] = name
		data["account"]["password"] = _response_["encrypted"].decode()
		data["permissions"]["activated"] = activated
		print("Saving data...")
		response = self.save_data(email=email, username=username, data=data)
		if not response.success: return response

		# insert api key cache.
		try: self.__api_keys__
		except AttributeError: self.__api_keys__ = {}
		self.__api_keys__[data["keys"]["api_key"]] = {
			"email":email,
			"username":username
		}

		# handle success.
		return dev0s.response.success(f"Successfully created user [{email}].", {
			"user":user,
			"email":user.email,
			"data":data,
		})

		#
	def update(self,
		# required:
		username=None,
		email=None,
		# optionals:
		name=None,
		password=None,
		verify_password=None,
		superuser=None,
		#phone_number=None,
		#photo_url=None,
		activated=None,
	):

		# checks.
		username, email = self.__correct_username_email__(username, email)

		# check parameters.
		if [username,email] == [None,None]:
			return dev0s.response.error(self.__traceback__(function="update")+" Define one of the following parameters [email, username].")

		# load.
		response = self.get(email=email, username=username)
		if response.error != None: return response
		user = response["user"]

		# check password.
		if password != None:
			if verify_password == None: return dev0s.response.error("Specify parameter: verify_password.")
			password = str(password)
			verify_password = str(verify_password)
			if len(password) < 8:
				return dev0s.response.error("The password must contain at least 8 characters.")
			elif password.lower() == password:
				return dev0s.response.error("The password must regular and capital letters.")
			elif password != verify_password:
				return dev0s.response.error("Passwords do not match.")
		
		# create django user.
		response = self.django.users.update(
			email=email,
			username=username,
			password=password,
			name=name,
			superuser=superuser,
			activated=activated,)
		if not response.success: return response

		# update firestore.
		response = self.load_data(email=email, username=username)
		if not response.success: return response
		data, edits = response.data, 0
		if email != None:
			edits += 1
			data["account"]["email"] = email
		if name != None:
			edits += 1
			data["account"]["name"] = name
		if isinstance(activated, (bool, Boolean)):
			edits += 1
			data["permissions"]["activated"] = bool(activated)
		if edits > 0:
			response = self.save_data(email=email, username=username, data=data)
			if not response.success: return response

		# save pass.
		if password != None:
			response = self.save_password(email=email, username=username, password=password)
			if not response.success: return response

		# handle success.
		return dev0s.response.success(f"Successfully updated user [{email}].")
		
		#
	def delete(self, 
		# the user's email.
		email=None,
		# the user's username.
		username=None,
	):
		
		# checks.
		username, email = self.__correct_username_email__(username, email)

		# check parameters.
		if [username,email] == [None,None]:
			return dev0s.response.error(self.__traceback__(function="delete")+" Define one of the following parameters [email, username].")

		# delete django user.
		response = self.django.users.delete(username=username, email=email)
		if not response.success: return response

		# delete firestore data.
		response = self.db.delete(self.__get_path__(email=email, username=username))
		if not response.success: return response

		# handle success.
		return dev0s.response.success(f"Successfully deleted user [{email}].")

		#
	def exists(self, 
		# one of the user id options is required.
		username=None,
		email=None,
		# the filter (django).
		filter="django",
	):
		username, email = self.__correct_username_email__(username, email)
		if filter == "django":
			if email != None:
				return DjangoUser.objects.filter(email=email).exists()
			elif username != None:
				return DjangoUser.objects.filter(username=username).exists()
		else: 
			raise ValueError(f"Invalid filter [{filter}], valid options: [django].")
	def authenticate(self,
		# the users username.
		username=None,
		# the users password.
		password=None,
		# the 2fa code.
		_2fa_code=None,
		# the 2fa enabled boolean.
		# leave None to use the self._2fa settings and specify to overwrite.
		_2fa=None,
		# the html for the verification code email (str).
		html="",
		# the request object.
		request=None,
	):

		# check parameters.
		response = dev0s.response.parameters.check(
			traceback=self.__traceback__(function="authenticate"),
			parameters={
				"username":username,
				"password":password,
			})
		if not response.success: return response

		# check first.
		response = self.django.users.authenticate(username=username, password=password, login=not _2fa, request=request)
		if not response.success: return response
		
		# check 2fa.		
		if _2fa == None: _2fa = self._2fa
		if _2fa == True:
			if _2fa_code == None: 

				# send code.
				response = self.send_code(username=username, ip=utils.get_client_ip(request), mode="authentication", html=html)
				if not response.success: return response

				# handler for sign in.
				return dev0s.response.error(f"Authentication verification code required.")

			# verify code.
			response = self.verify_code(username, code=_2fa_code, mode="authentication")
			if not response.success: return response

		# no 2fa.
		else:
			if response.success:
				_response_ = self.get_api_key(username=username)
				if not _response_.success: return _response_
				response["api_key"] = _response_.api_key
				_response_ = self.load_data(username=username)
				if not _response_.success: return _response_
				response["name"] = _response_.data["account"]["name"]
				response["username"] = _response_.data["account"]["username"]
				response["email"] = _response_.data["account"]["email"]
				response["activated"] = _response_.data["permissions"]["activated"]
			return response

		# success.
		response = self.django.users.authenticate(username=username, password=password, request=request)
		if response.success:
			_response_ = self.get_api_key(username=username)
			if not _response_.success: return _response_
			response["api_key"] = _response_.api_key
			_response_ = self.load_data(username=username)
			if not _response_.success: return _response_
			response["name"] = _response_.data["account"]["name"]
			response["username"] = _response_.data["account"]["username"]
			response["email"] = _response_.data["account"]["email"]
			response["activated"] = _response_.data["permissions"]["activated"]
		return response

		#
	def signout(self,
		# the request object (obj) (#1).
		request=None,
	):
		if request.user in [None] or request.user.username in [None, ""]:
			return dev0s.response.error("There is no user signed in.")
		try:
			_logout_(request)
		except Exception as e:  return dev0s.response.error(f"Failed to sign out user [{request.user.username}], error: {e}.")
		return dev0s.response.success("Successfully signed out.")
		#
	def authenticated(self, 
		# the request (#1).
		request=None,
	):

		# by request user filled (django).
		if request.user.is_authenticated:
			return dev0s.response.success(f"User {request.user.username} is authenticated.", {
				"api_key":None,
				"email":request.user.email,
				"username":request.user.username,
			})

		# by api key.
		optional_parameters, _ = dev0s.response.parameters.get(request, {
			"api_key":None,
		})
		if optional_parameters["api_key"] != None:
			return self.verify_api_key(api_key=optional_parameters["api_key"])

		# error.
		return dev0s.response.error("User is not authenticated, please specify your api key or login.")

		#
	def root_permission(self, 
		# the request (#1).
		request=None,
	):

		# by request user filled (django).
		if request.user != None and request.user.is_authenticated and request.user.is_superuser:
			return dev0s.response.success(f"Root permission granted ({request.user.username}).", {
				"api_key":None,
				"email":request.user.email,
				"username":request.user.username,
			})

		# by api key.
		# never root permission.

		# error.
		return dev0s.response.error("Root permission denied.")

		#
	def load_data(self,
		# the user's email.
		email=None,
		# the user's username.
		username=None,
		# the create boolean (do not use).
		create=False,
	):	
		username, email = self.__correct_username_email__(username, email)
		if [username,email] == [None,None]:
			return dev0s.response.error(self.__traceback__(function="load_data")+" Define one of the following parameters [email, username].")
		response = self.db.load(path=self.__get_path__(email=email, username=username, create=create))
		if response.error != None: return response
		return dev0s.response.success(f"Successfully loaded the data of user [{email}].", {
			"data":response["data"],
		})
	def save_data(self,
		# the user's email.
		email=None,
		# the user's username.
		username=None,
		# the user's data.
		data={},
		# the overwrite boolean.
		overwrite=False,
	):
		username, email = self.__correct_username_email__(username, email)
		if [username,email] == [None,None]:
			return dev0s.response.error(self.__traceback__(function="save_data")+" Define one of the following parameters [email, username].")
		response = self.db.save(path=self.__get_path__(email=email, username=username, create=True), data=data, overwrite=overwrite)
		if response.error != None: return response
		return dev0s.response.success(f"Successfully saved the data of user [{email}].")
	def send_email(self, 
		# define email to retrieve user.
		email=None, 
		username=None,
		# the email title.
		title="Account Activation",
		# the html (str).
		html="",
	):

		# check email.
		if self.email.__class__.__name__ not in ["Email"]:
			response = self.__initialize_email__()
			if response.error != None: return response

		# check username email.
		username, email = self.__correct_username_email__(username, email)
				
		# get user.
		response = self.get(email=email, username=username)  
		if response.error != None: return response
		user = response["user"]
		
		# send email.
		if dev0s.defaults.options.log_level >= 1:
			dev0s.response.log(f"Sending email to user [{user.username}], email [{user.email}].")
		response = self.email.send(
			subject=title,
			recipients=[user.email], 
			html=html)
		if response.error != None:
			response = self.__initialize_email__()
			if response.error != None: return response
			response = self.email.send(
				subject=title,
				recipients=[user.email], 
				html=html)
			if response.error != None: return response

		# success.
		return dev0s.response.success(f"Successfully send an email to user [{user.email}].")

		#
	def send_code(self, 
		# define username / email to retrieve user.
		username=None, 
		email=None,
		# the clients ip.
		ip="unknown",
		# the mode id.
		mode="verification",
		# the mode title.
		title="Account Activation",
		# the html (str).
		html="",
		# optionally specify the code (leave None to generate).
		code=None,
	):

		# check email.
		if self.email.__class__.__name__ not in ["Email"]:
			response = self.__initialize_email__()
			if response.error != None: return response
		
		# check username email.
		username, email = self.__correct_username_email__(username, email)

		# save code.
		if code == None:
			code = Integer(0).generate(length=6)
		response = self.get(username=username, email=email)
		if response.error != None: return response
		user = response["user"]
		self.__verification_codes__.load()
		try: self.__verification_codes__[mode]
		except: self.__verification_codes__[mode] = {}
		self.__verification_codes__[mode][user.email] = {
			"code":code,
			"stamp":Date().timestamp,
			"attempts":3,
		}
		self.__verification_codes__.save()

		# html vars.
		domain = utils.utils.naked_url(self.domain)
		if dev0s.system.env.get("PRODUCTION", format=bool, default=True):
			domain = "https://"+utils.utils.naked_url(self.domain)
		for from_, to_ in {
			"$DOMAIN": domain,
			"$USERNAME": user.username,
			"$EMAIL": user.email,
			"$CODE": code,
			"$IP": ip,
			"$TIMESTAMP": Date().timestamp,
		}.items():
			html = html.replace(f"({from_})", to_)

		# html colors.
		for key,value in self.template_data["COLORS"].items():
			html = html.replace(f"(${key.upper()})", str(value))

		# send email.
		if dev0s.defaults.options.log_level >= 1:
			dev0s.response.log(f"Sending [{mode}] code to user [{user.username}], email [{user.email}].")
		response = self.email.send(
			subject=f'{title} - Verification Code',
			recipients=[user.email], 
			html=html)
		if response.error != None:
			response = self.__initialize_email__()
			if response.error != None: return response
			response = self.email.send(
				subject=f'{title} - Verification Code',
				recipients=[user.email], 
				html=html)
			if response.error != None: return response

		# success.
		return dev0s.response.success(f"Successfully send a verification code to user [{user.email}].")

		#
	def verify_code(self, 
		# define email to retrieve user.
		username=None, 
		email=None,
		# the user entered code.
		code=000000, 
		# the message mode.
		mode="verification",
	):


		# check username email.
		username, email = self.__correct_username_email__(username, email)

		# get user.
		response = self.get(username=username, email=email)
		if response.error != None: return response
		user = response["user"]

		# get success.
		self.__verification_codes__.load()
		fail = False
		try: self.__verification_codes__[mode]
		except: self.__verification_codes__[mode] = {}
		try: 
			date = Date()
			decreased = date.decrease(date.timestamp, format=date.timestamp_format, minutes=5)
			outdated = Date(self.__verification_codes__[mode][user.email]["stamp"]) <= Date(decreased)
			success = not outdated and self.__verification_codes__[mode][user.email]["attempts"] > 0 and str(self.__verification_codes__[mode][user.email]["code"]) == str(code)
		except KeyError: fail = True
		# handle.
		if fail: 
			return dev0s.response.error("Incorrect verification code.")
		elif not success: 
			self.__verification_codes__[mode][user.email]["attempts"] -= 1
			self.__verification_codes__.save()
			return dev0s.response.error("Incorrect verification code.")
		else:
			del self.__verification_codes__[mode][user.email]
			self.__verification_codes__.save()
			return dev0s.response.success("Successfull verification.")

		#
	def verify_api_key(self, api_key=None, request=None):
		
		# by request.
		if request != None:
			parameters, response = dev0s.response.parameters.get(request, ["api_key"])
			if not response.success: return response
			return self.verify_api_key(api_key=parameters["api_key"])
		
		# by api key.
		else:
			
			# collect api keys from cache.
			response = self.__collect_api_keys_cache__()
			if not response.success: return response
			api_keys = response["api_keys"]

			# check api key.
			for _api_key_, info in api_keys.items():
				if _api_key_ == api_key: 
					return dev0s.response.success(f"Successfully verified api key [{api_key}].", {
						"email":info["email"],
						"username":info["username"],
						"api_key":api_key,
					})

			# error.
			return dev0s.response.error(f"Unable to verify api key [{api_key}].")

			#
	def verify_subscription(self,
		# select one of the following user id options:
		email=None,
		username=None,
		api_key=None,
		# the subscription product.
		product=None,
		# the subscription plans that will return a success verification (["*"] for all plans within the product).
		plans=[],
	):

		# check username email.
		username, email = self.__correct_username_email__(username, email)

		# set user id.
		if email != None:
			id = email
		elif username != None:
			email = self.__username_to_email__(username)
			id = email
		elif api_key != None:
			response = self.verify_api_key(api_key=api_key)
			if not response.success: return response
			id = response["email"]
			email = response["email"]

		# collect stripe subscriptions cache.
		response = self.__collect_subscriptions_cache__()
		if not response.success: return response
		subscriptions = response["subscriptions"]

		# create user subscriptions.
		user_subscriptions = {}
		for _product_, _plans_ in self.stripe._subscriptions_.items():
			response = self.stripe.get_product_id(product=_product_)
			if not response.success: return response
			product_id = response["id"]
			user_subscriptions[_product_] = {}
			for plan_name, plan_info in _plans_.items():
				response = self.stripe.get_plan_id(product=_product_, plan=plan_name)
				if not response.success: return response
				plan_id = response["id"]
				active = False
				try: _subscriptions_ = subscriptions[email]
				except: _subscriptions_ = {}
				for _plan_id_, info in _subscriptions_.items():
					if _plan_id_ == plan_id and info["active"]:
						active = True
						break
				user_subscriptions[_product_][plan_name] = active

		# check user subscriptions.
		required_plans = list(plans)
		verified = False
		for allowed_plan in plans:
			if allowed_plan == "*":
				required_plans = list(user_subscriptions[product].keys())
				for _, allowed in user_subscriptions[product].items():
					if allowed:
						verified = True
						break
				break
			else:
				if user_subscriptions[product][allowed_plan] == True:
					verified = True
					break

		# handlers.
		if verified:
			return dev0s.response.success(f"Successfully verified subscription {product} for user {id}.", {
				"email":email,
				"api_key":api_key,
				"product":product,
				"plans":required_plans,
			})
		else:
			_required_plans_ = ""
			for i in required_plans: 
				if _required_plans_ == "": _required_plans_ = i
				else: _required_plans_ += ", "+i
			return dev0s.response.error(f"Permission denied ({email}). This request requires at least one of the following plans for {product} [{_required_plans_}].")

		#
	def create_subscription(self,
		# select one of the following user id options:
		email=None,
		username=None,
		api_key=None,
		# the subscription product.
		product=None,
		# the subscription plan.
		plan=None,
		# the card holders name.
		card_name=None,
		# the card number.
		card_number=None,
		# the card expiration month.
		card_expiration_month=None,
		# the card expiration year.
		card_expiration_year=None,
		# the card cvc.
		card_cvc=None,
	):

		# check username email.
		username, email = self.__correct_username_email__(username, email)

		# checkparameters.
		if username == None and email == None and api_key == None: return dev0s.response.error("Define one of the following parameters: [email, api_key].")
		response = dev0s.response.parameters.check(
			traceback=self.__traceback__(function="create_subscription"),
			parameters={
				"product":product,
				"plan":plan,
			})
		if not response.success: return response
		blinded = False
		if card_cvc == "***" and "************" in card_number:
			blinded = True
		if not blinded:
			response = dev0s.response.parameters.check(
				traceback=self.__traceback__(function="create_subscription"),
				parameters={
					"card_name":card_name,
					"card_number":card_number,
					"card_expiration_month":card_expiration_month,
					"card_expiration_year":card_expiration_year,
					"card_cvc":card_cvc,
				})
			if not response.success: return response
		

		# set user id.
		if email != None:
			id = email
		elif username != None:
			email = self.__username_to_email__(username)
			id = email
		elif api_key != None:
			response = self.verify_api_key(api_key=api_key)
			if not response.success: return response
			id = response.email

		# get plan id.
		response = self.stripe.get_product_id(product=product)
		if not response.success: return response
		product_id = response.id
		response = self.stripe.get_plan_id(product=product, plan=plan)
		if not response.success: return response
		plan_id = response.id

		# check / create customer.
		response = self.stripe.customers.check(email=id)
		if not response.success: return response
		elif not response.exists:
			response = self.stripe.customers.create(email=id)
			if not response.success: return response
			customer_id = response.id
		else:
			customer_id = response.id

		# map user's stripe subscriptions.
		response = self.stripe.subscriptions.get(email=email)
		if not response.success: return response
		plan_ids, sub_ids, active_product_subs = [], {}, {}
		for _plan_id_,info in response.subscriptions.items():
			response = self.stripe.get_product_id_by_plan_id(_plan_id_)
			if not response.success: return response
			_product_id_ = response.id
			if _product_id_ == product_id:
				active_product_subs[info["plan_id"]] = info["subscription_id"]	
			plan_ids.append(info["plan_id"])
			sub_ids[info["plan_id"]] = info["subscription_id"]

		# check already subscribed.
		if plan_id in plan_ids:
			return dev0s.response.error(f"User {email} is already subscribed to plan {plan} from product {product}.")

		# create card when card is not blinded (aka retrieved from user).
		if not blinded:
			response = self.stripe.customers.create_card(
				id=customer_id,
				name=card_name,
				number=card_number,
				month=card_expiration_month,
				year=card_expiration_year,
				cvc=card_cvc,)
			if not response.success: return response

		# cancel all plans within the product.	
		for _plan_id_, _subscription_id_ in active_product_subs.items():
			response = self.stripe.subscriptions.cancel(subscription_id=_subscription_id_)
			if not response.success: return response		

		# create subscription.
		response = self.stripe.subscriptions.create(customer_id=customer_id, plans=[plan_id])
		if not response.success: return response
		
		# handlers.
		return dev0s.response.success(f"Successfully created subscription {product} {plan} for user {id}.")

		#
	def get_api_key(self, email=None, username=None):

		# check.
		if [username,email] == [None,None]:
			return dev0s.response.error("Define one of the following parameters: [email, username].")
		if username == None: id = email
		elif email == None: id = username

		# collect api keys from cache.
		response = self.__collect_api_keys_cache__()
		if not response.success: return response
		api_keys = response["api_keys"]

		# get api key.
		api_key = None
		for _api_key_, info in api_keys.items():
			if (email != None and email == info["email"]) or (username != None and username == info["username"]): 
				api_key = _api_key_
				break
		if api_key != None:
			return dev0s.response.success(f"Successfully retrieved the api key [{id}].", {
				"api_key":api_key,
			})
		else:
			return dev0s.response.error(f"Unable to find api key [{id}].")

		#
	def set_permission(self, email=None, username=None, permission_id=None, permission=True):
		
		# load user data.
		response = self.load_data(email=email, username=username) 
		if not response.success: return response
		data = response["data"]

		# edit data.
		data["permissions"][permission_id] = permission

		# save data.
		response = self.save_data(email=email, username=username, data=data) 
		if response.error != None: return response

		# handlers.
		return dev0s.response.success(f"Successfully set the {permission_id} permission of user [{email}] to [{permission}].")
		#

	# user passwords.
	def check_password(self, 
		# password (#1).
		password=None, 
		# verify password (#2).
		verify_password=None, 
		# the strong password boolean.
		strong=False,
	):
		response =dev0s.response.parameters.check(
			traceback=self.__traceback__(function="check_password"),
			parameters={
				"password:str,String":password,
				"verify_password:str,String":verify_password,
				"strong:bool,Boolean":strong,
			})
		if not response.success: return response
		if password != verify_password:
			return dev0s.response.error("Passwords do not match.")
		elif (not strong and len(password) < 6) or (strong and len(password) < 12):
			return dev0s.response.error("The password must contain at least 12 characters.")
		elif password.lower() == password:
			return dev0s.response.error("The password must contain capital letters.")
		match = False
		for i in [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]:
			if  str(i) in password: match = True ; break
		if match == False:
			return dev0s.response.error("The password must contain digits.")
		return dev0s.response.success("Succesfully checked the password.")
	def load_password(self, email=None, username=None):
		username, email = self.__correct_username_email__(username, email)
		if [username,email] == [None,None]:
			return dev0s.response.error(self.__traceback__(function="load_password")+" Define one of the following parameters [email, username].")
		response = self.load_data(email=email, username=username)
		if not response.success: return response
		data = response.data
		_response_ = self.aes.decrypt(data["account"]["password"])
		if not _response_.success: return _response_
		password = _response_["decrypted"].decode()
		return dev0s.response.success(f"Successfully retrieved the password of user {email}.",{
			"password":password,
			"data":data,
		})
	def save_password(self, email=None, username=None, password=None):
		username, email = self.__correct_username_email__(username, email)
		if [username,email] == [None,None]:
			return dev0s.response.error(self.__traceback__(function="save_password")+" Define one of the following parameters [email, username].")
		response = dev0s.response.parameters.check(
			traceback=self.__traceback__(function="save_password"),
			parameters={
				"password":password,
			})
		if not response.success: return response
		response = self.load_data(email=email, username=username)
		if not response.success: return response
		data = response.data
		_response_ = self.aes.encrypt(password)
		if not _response_.success: _response_
		try:data["account"]
		except: data["account"] = {}
		data["account"]["password"] = _response_["encrypted"].decode()
		response = self.save_data(email=email, username=username, data=data)
		if not response.success: return response
		return dev0s.response.success(f"Successfully saved the password of user {email}.")

	# iterations.
	def iterate(self,
		# the filter of what to iterate.
		filter="user",
		# from which database to iterate (django / database).
		database="database",
	):

		# vars.
		database_options = ["django", "database"]
		filter_options = ["user", "username", "email", "data"]

		# normalize.
		filter = filter.lower()
		database = database.lower()
		if database in ["db"]: database = "database"
		if database not in database_options and database[:-1] in database_options: database = database[:-1]

		# checks.
		if database not in database_options:
			raise dev0s.exceptions.InvalidUsage(f"{self.__traceback__(function='iterate',parameter='database')}: Selected an invalid database option [{database}], valid options: [{Array(database_options).string(joiner=', ')}].")
		if filter not in filter_options:
			raise dev0s.exceptions.InvalidUsage(f"{self.__traceback__(function='iterate',parameter='filter')}: Selected an invalid filter option [{filter}], valid options: [{Array(filter_options).string(joiner=', ')}].")

		

		# database: django.
		items = []
		if database == "django":
			users = auth.list_users().iterate_all()
			for user in users:
				if filter == "user":
					items.append(user)
				elif filter == "email":
					items.append(user.email)
				elif filter == "username":
					items.append(user.username)
				elif filter == "data":
					response = self.load_data(email=user.email, username=user.username)
					if not response.success: 
						dev0s.response.log(f"Unable to iterate user: [{user.username}]  (#346346) (error: {response.error}).")
					else:
						items.append(response["data"])

		# database: database.
		elif database == "database":

			# firestore mode.
			for id in self.db.names(path=self.users_subpath):
				username, email = self.__correct_username_email__(id, id)
				if username != None: id = username
				elif email != None: id = email
				response = self.load_data(email=email, username=username)
				if not response.success: 
					dev0s.response.log(f"Unable to iterate user: [{id}] (#837028) (error: {response.error}).")
				else:
					username, email = response["data"]["account"]["username"], response["data"]["account"]["email"]
					if filter == "user":
						response = self.django.users.get(email=email, username=username)
						if not response.success: 
							dev0s.response.log(f"Unable to iterate (django) user: [{username}].")
						else:
							items.append(response["user"])
					elif filter == "email":
						items.append(email)
					elif filter == "username":
						items.append(username)
					elif filter == "data":
						items.append(response["data"])

		# handler.
		return items

		#
	def synchronize(self, 
		# leave ids=None default to synchronize all users.
		# optionally pass emails=[newuser@email.com] to synchronize new users.
		emails=["*"], 
		usernames=["*"], 
	):

		# get all ids.
		_usernames_ = []
		if "*" in emails and "*" in usernames:
			_usernames_ = self.iterate(database="database", filter="username")
		elif "*" not in emails:
			for email in emails:
				response = self.django.users.get(email=email)
				if not response.success:
					return response
				else:
					_usernames_.append(response.user.username)
		elif "*" not in usernames:
			usernames_ = usernames
		else:
			_usernames_ = usernames
			for email in emails:
				response = self.django.users.get(email=email)
				if not response.success:
					return response
				else:
					_usernames_.append(response.user.username)

		# iterate.
		api_keys = {}
		for username in _usernames_: 

			# load data.
			response = self.load_data(username=username)
			if not response.success: return response
			data, edits = response["data"], 0

			# check user data.
			d = Dictionary(dictionary=data, path=False)
			old = dict(data)
			data = d.check(default=dict(self.default_user_data))
			if Dictionary(old) != Dictionary(data):
				edits += 1

			# safe early edits for other funcs.
			if edits > 0:
				edits = 0
				response = self.save_data(username=username, data=data)
				if not response.success: return response

			# get.
			response = self.get(email=None, username=username)
			if not response.success: 
				response = self.load_password(email=None, username=username)
				if not response.success: return response
				password = response["password"]
				response = self.django.users.create(
					email=data["account"]["email"], 
					password=password, 
					username=data["account"]["username"], 
					name=data["account"]["name"])
				if not response.success: return response
			user = response["user"]


			# check user data.
			try:
				if data["keys"]["api_key"] in [None, ""]: raise KeyError("")
			except KeyError: 
				data["keys"]["api_key"] = String().generate(length=48, capitalize=True, digits=True, special=False).replace("+","")
				edits += 1

			# edits.
			if edits > 0:
				response = self.save_data(email=user.email, username=user.username, data=data)
				if not response.success: return response

			# api keys.
			api_keys[data["keys"]["api_key"]] = {
				"username":user.username,
				"email":user.email,
			}
			
		# success.
		return dev0s.response.success(f"Successfully synchronized {len(_usernames_)} user(s).", {
			"api_keys":api_keys,
		})

		#

		#



	# system functions.
	def __initialize_email__(self):
		# the email object.
		if not self.email_enabled:
			return dev0s.response.error("The email object is disabled.")
		if self.email_address == None or self.email_password == None:
			return dev0s.response.error("Define the firebase.users.email_address & firebase.users.email_password variables to send emails.")
		self.email = email.Email(
			email=self.email_address,
			password=self.email_password,
			smtp_host=self.smtp_host,
			smtp_port=self.smtp_port,
			visible_email=self.visible_email,)
		response = self.email.login()
		if response["success"]:
			return dev0s.response.success("Successfully initialized the mail object.")
		else: 
			self.email = None
			return response
	def __collect_subscriptions_cache__(self, refresh=False):

		# check cache.
		mode = "subscriptions"
		try: self.__timestamps__[mode]
		except: 
			self.__timestamps__[mode] = Date()
			refresh = True
		if not refresh:
			date = Date()
			decreased = date.decrease(date.timestamp, format=date.timestamp_format, hours=24)
			if self.__timestamps__[mode] <= Date(decreased):
				refresh = True

		# get cache.
		subscriptions = {}
		if not refresh:
			subscriptions = dict(self.__subscriptions__)

		# collect cache.
		if refresh:
			response = self.stripe.subscriptions.get(active_only=True, by_customer_id=False)
			if not response.success: return response
			subscriptions = response["subscriptions"]

		# set cache.
		if refresh:
			self.__subscriptions__ = dict(subscriptions)
			self.__timestamps__[mode] = Date()

		# handler.
		return dev0s.response.success("Successfully retrieved the stripe subscriptions cache.", {
			"subscriptions":subscriptions,
		})

		#
	def __collect_api_keys_cache__(self, refresh=False):
		
		# check cache.
		mode = "api_keys"
		try: self.__timestamps__[mode]
		except: 
			self.__timestamps__[mode] = Date()
			refresh = True
		if not refresh:
			date = Date()
			decreased = date.decrease(date.timestamp, format=date.timestamp_format, hours=24)
			if self.__timestamps__[mode] <= Date(decreased):
				refresh = True

		# get cache.
		api_keys = {}
		if not refresh:
			api_keys = dict(self.__api_keys__)

		# collect cache.
		if refresh:
			response = self.synchronize()
			if not response.success: return response
			api_keys = response["api_keys"]

		# set cache.
		if refresh:
			self.__api_keys__ = dict(api_keys)
			self.__timestamps__[mode] = Date()

		# handler.
		return dev0s.response.success("Successfully retrieved the api keys cache.", {
			"api_keys":api_keys,
		})

		#
	def __get_path__(self, email=None, username=None, create=False):
		username, email = self.__correct_username_email__(username, email)
		if self.id_by_username:
			if username == None and email != None:
				response = self.get(email=email)
				if not response.success: response.crash()
				username = response.user.username
			id = username
			if "@" in username:

				raise ValueError(f"Forbidden @ character found in username: [{username}].")
		else:
			if email == None and username != None:
				response = self.get(username=username)
				if not response.success: response.crash()
				email = response.user.email
			id = email
			if "@" not in email:
				raise ValueError(f"Required @ character not found in email: [{email}].")
		if self.db.mode == "cache":
			path = f"{self.users_subpath}/{id}/settings"
			if create and not Files.exists(gfp.base(path)): Files.create(gfp.base(path), directory=True)
		else:
			path = f"{self.users_subpath}/{id}"
		return path
	def __username_to_email__(self, username):
		try:
			return self.get(username=username).email
		except Exception as e:
			raise ValueError(f"Unable retrieve the email of username [{username}], error: {e}.")
	def __email_to_username__(self, email):
		try:
			return self.get(email=email).username
		except Exception as e:
			raise ValueError(f"Unable retrieve the username of email [{email}], error: {e}.")
	def __correct_username_email__(self, username, email):
		if username in ["None", "null", "NaN"]: username = None
		if email in ["None", "null", "NaN"]: email = None
		if email in [None, username] and (username != None and "@" in username): 
			email = username 
			username = None
		if username in [None, email] and (email != None and "@" not in email): 
			username = email
			email = None
		return username, email
	#

#




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# alias.
ALIAS = "vbackups"

# updates.
import os, sys
from dev0s.shortcuts import *
if dev0s.cli.arguments_present(["--update"]) and ALIAS in sys.argv[0]:
	os.system(f"curl -s https://raw.githubusercontent.com/vandenberghinc/{ALIAS}/master/requirements/installer.remote | bash ")
	sys.exit(0)

# imports.
import os, sys, ast, requests, json, glob, subprocess, random, platform, time, socket, time, threading, urllib
from datetime import datetime

# inc imports.
from dev0s.shortcuts import *
from inc_package_manager import inc_package_manager

# source.	
SOURCE = Directory(path=dev0s.defaults.source_path(__file__, back=3))

# checks.
dev0s.defaults.operating_system(supported=["linux", "macos"])
dev0s.defaults.alias(alias=ALIAS, executable=SOURCE.fp.path, venv=SOURCE.join("venv"))

# env settings.
PRODUCTION = dev0s.env.get_boolean("PRODUCTION", default=True)
DAEMON_PRODUCTION = dev0s.env.get_boolean("DAEMON_PRODUCTION", default=True)
MIGRATIONS = dev0s.env.get_boolean("MIGRATIONS", default=False)

# universal variables.
ADMINISTRATOR = "administrator"

# database.
DATABASE = Directory(path=dev0s.env.get_string("VBACKUPS_DATABASE", default=f"{dev0s.defaults.vars.home}/.{ALIAS}"))
if not DATABASE.fp.exists():
	dev0s.response.log(f"{color.orange}Root permission{color.end} required to create {ALIAS} database [{DATABASE}].")
	os.system(f" sudo mkdir -p {DATABASE}")
	Files.chown(str(DATABASE), owner=dev0s.defaults.vars.owner, group=dev0s.defaults.vars.group, sudo=True, recursive=True)
	Files.chmod(str(DATABASE), permission=700, sudo=True, recursive=True)

# ssht00ls.
for dir in [
	SOURCE.join("__defaults__/env"),
	DATABASE.join("keys/ssht00ls"),
]:
	if not Files.exists(dir): Files.create(dir, directory=True)
dev0s.system.env.export(export=[SOURCE.join("__defaults__/env/bash"), SOURCE.join("__defaults__/env/json")], env={
	"INTERACTIVE":False,
	"SSHT00LS_CONFIG":DATABASE.join("keys/ssht00ls/config"),
	"SSHT00LS_PRIVATE_KEY":DATABASE.join("keys/ssht00ls/private_key"),
	"SSHT00LS_PUBLIC_KEY":DATABASE.join("keys/ssht00ls/public_key"),
})
import ssht00ls, w3bsite
from ssht00ls import ssht00ls_agent

# w3bsite.

# variables.
LOG_FILE = f"{DATABASE}/logs/logs"
ERROR_FILE = f"{DATABASE}/logs/errors"
dev0s.response.log_file = LOG_FILE

# config.
CONFIG = Dictionary(path=DATABASE.join("config",""), load=True, default={
})
edits = 0
if edits > 0: CONFIG.save()

# limit migrations.
if not MIGRATIONS:

	# check db.
	SUDO, PERMISSION = True, 770
	DATABASE.check(
		owner=dev0s.defaults.vars.user,
		group=dev0s.defaults.vars.group,
		permission=PERMISSION,
		sudo=SUDO,
		recursive=False,
		silent=False,
		hierarchy={
			"data/":{
				"path":"data/",
				"directory":True,
				"owner":dev0s.defaults.vars.user,
				"group":dev0s.defaults.vars.group,
				"permission":PERMISSION,
				"sudo":SUDO,
			},
			"backups/":{
				"path":"backups/",
				"directory":True,
				"owner":dev0s.defaults.vars.user,
				"group":dev0s.defaults.vars.group,
				"permission":PERMISSION,
				"sudo":SUDO,
			},
			"clients/":{
				"path":"clients/",
				"directory":True,
				"owner":dev0s.defaults.vars.user,
				"group":dev0s.defaults.vars.group,
				"permission":PERMISSION,
				"sudo":SUDO,
			},
			f"clients/{ADMINISTRATOR}/":{
				"path":f"clients/{ADMINISTRATOR}/",
				"directory":True,
				"owner":dev0s.defaults.vars.user,
				"group":dev0s.defaults.vars.group,
				"permission":PERMISSION,
				"sudo":SUDO,
			},
			"logs/":{
				"path":"logs/",
				"directory":True,
				"owner":dev0s.defaults.vars.user,
				"group":dev0s.defaults.vars.group,
				"permission":PERMISSION,
				"sudo":SUDO,
			},
		})

	# website.
	website = w3bsite.Website(
		# main
		root=SOURCE.fp.path,
		database=DATABASE.fp.path,
		library=f"/usr/local/lib/{ALIAS}",
		domain="0.0.0.0:8002",
		name="VBackups",
		remote=None,
		# the template data (only required when running the website) (overwrites the w3bsite template data keys).
		template_data={
			# customize colors.
			"COLORS":{
				"white":"#FAFAFA",
				"light_white":"#E9F0FD",
				"grey":"#E5E5E5",
				"light_grey":"#D6D6D6",
				"dark_grey":"#424242",
				"blue":"#5A8FE6",
				"purple":"#323B83",#"#B32FCA",
				#"purple":"#9B00AA",
				"red":"#FD304E",
				"pink":"#F62B7D",
				"orange":"#FF8800",
				"green":"#006633",
				"darkest":"#1F2227",
				"darker": "#20242A",
				"dark": "#262B30",
				# background color.
				"topbar":"#FAFAFA",#1F2227", #"#FAFAFA",
				"background":"#FAFAFA",#"#E7E9EF", #"#FAFAFA",
				"topbar_darkmode":"#1F2227",#1F2227", #"#FAFAFA",
				"background_darkmode":"#1F2227",#"#E7E9EF", #"#FAFAFA",
				# elements.
				"widgets":"#FAFAFA",
				"widgets_reversed":"#323B83",#"#1F2227",
				"widgets_darkmode":"#20242A",
				"widgets_reversed_darkmode":"#323B83",#"#1F2227",
				# text.
				"text":"#1F2227",
				"text_reversed":"#FAFAFA",
				"text_darkmode":"#FAFAFA",
				"text_reversed_darkmode":"#FAFAFA",
				# input & textareas.
				"input_txt":"#6C6B6D",
				"input_txt_reversed":"#FAFAFA",
				"input_bg":"#E9F0FD", #"#FAFAFA", 
				"input_bg_reversed":"#323B83",
				
				# buttons.
				"button_txt":"#FAFAFA",
				"button_txt_reversed":"#1F2227",
				"button_bg":"#323B83",
				"button_bg_reversed":"#FAFAFA",
				# custom colors.
				# ...
			},
		},
		# custom styling.
		styling={
			"LEFTBAR_WIDTH":"280px", # px
			"RIGHTBAR_WIDTH":"280px", # px
			"TOPBAR_HEIGHT":"50px", #px
		},
		# 	the organization name.
		organization="VanDenBerghInc",
		# aes.
		aes_passphrase=None,
		# options.
		namecheap_enabled=False,
		firebase_enabled=False,
		stripe_enabled=False,
		email_enabled=False,
		users_subpath="clients/",
		id_by_username=True,
		interactive=False,
		_2fa=False,
		production=PRODUCTION,
	)

	# export env.
	website.security.set_secret_env("VBACKUPS_DATABASE", str(DATABASE))
		
	#


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes import utils, views
from w3bsite.classes import defaults as _defaults_

# the logging requests.
class Requests(_defaults_.Defaults):
	def __init__(self, 
		# passed Website.x objects.
		defaults=None,
	):
		# defaults.
		_defaults_.Defaults.__init__(self)
		self.assign(defaults.dict())

		# urlpatterns.
		if dev0s.system.env.get("MIGRATIONS", format=bool, default=False):
			self.urls = []
		else:
			self.urls = views.build_urls([
				self.Load(defaults=defaults),
				self.Reset(defaults=defaults),
			])

	# load the logs.
	class Load(views.Request):
		def __init__(self, defaults=None):
			views.Request.__init__(self, "requests/logs/", "load", website=defaults.website)
			self.assign(defaults.dict())
		@views.method_decorator(views.login_required)
		def request(self, request):

			# check root permissions.
			response = self.users.authenticated(request)
			if not response.success: return response

			# check root permissions.
			response = self.users.root_permission(request)
			if not response.success: return response

			# make request.
			return dev0s.response.load_logs(format="webserver")

			#

	# reset logs.
	class Reset(views.Request):
		def __init__(self, defaults=None):
			views.Request.__init__(self, "requests/logs/", "reset", website=defaults.website)
			self.assign(defaults.dict())
		@views.method_decorator(views.login_required)
		def request(self, request):

			# check root permissions.
			response = self.users.authenticated(request)
			if not response.success: return response

			# check root permissions.
			response = self.users.root_permission(request)
			if not response.success: return response

			# make request.
			return dev0s.response.reset_logs(format="webserver")

			#

	#

#



# imports
from w3bsite.classes.views.defaults import *

# the products overview.
class Products(View):
	def __init__(self, 
		# the base path (required; if url path is null) [#1 argument].
		base=None,
		# the views id (required) [#2 argument].
		id=None, 
		# the url path (optional).
		url=None,
		# the html path (optional).
		html=None,
		# enable if this view is the [/] landing page.
		landing_page=False,
		# the template data (required).
		template_data={},
		# the w3bsite.Website object.
		website=None,
	):

		# docs.
		DOCS = {
			"module":"website.views.Products", 
			"initialized":False,
			"description":[], 
			"chapter": "Views", }

		# defaults.
		View.__init__(self,
			base=base,
			id=id,
			url=url,
			html=html,
			landing_page=landing_page,
			template_data=template_data,
			# the view type.
			type="DocumentationView",)
		self.website = website
	def view(self, request):
		return self.render(request, self.template_data, html=f"{SOURCE_PATH}/classes/views/html/products.hml")


# the product view.
class Product(View):
	def __init__(self, 
		# the base path (required; if url path is null) [#1 argument].
		base=None,
		# the views id (required) [#2 argument].
		id=None, 
		# the url path (optional).
		url=None,
		# the html path (optional).
		html=None,
		# enable if this view is the [/] landing page.
		landing_page=False,
		# the template data (required).
		template_data={},
		# the w3bsite.Website object.
		website=None,
	):
		# docs.
		DOCS = {
			"module":"website.views.Product", 
			"initialized":False,
			"description":[], 
			"chapter": "Views", }

		# defaults.
		View.__init__(self,
			base=base,
			id=id,
			url=url,
			html=html,
			landing_page=landing_page,
			template_data=template_data,
			# the view type.
			type="DocumentationView",)
		self.website = website
	def view(self, request):
		return self.render(request, self.template_data, html=f"{SOURCE_PATH}/classes/views/html/product.hml")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes.utils import utils

# django imports.
from django.shortcuts import render
from django.http import HttpResponse, HttpResponseNotAllowed, JsonResponse
from django.urls import path, include
from django.contrib.auth.decorators import login_required
from django.utils.decorators import method_decorator
from django.views.decorators.clickjacking import xframe_options_exempt

# include apps from website/urls.py.
def include_apps(apps=[], auto_include=False):
	if dev0s.system.env.get("MIGRATIONS", format=bool, default=False):
		return []
	if auto_include:
		apps = []
		for _name_ in os.listdir(f"apps/"):
			if _name_ not in [".DS_Store"] and os.path.isdir(f"apps/{_name_}"): apps.append(_name_)
	urlpatterns = []
	for name in apps:
		_path_ = f"apps/{name}/"
		#try:
		if Files.exists(f"{_path_}/views.py"):
			l_path = f"{_path_}/views.py"
			urlpatterns.append(
				path('', include(f"apps.{name}.views"))
			)
		if Files.exists(f"{_path_}/requests.py"):
			l_path = f"{_path_}/requests.py"
			urlpatterns.append(
				path('', include(f"apps.{name}.requests"))
			)
		"""
		except Exception as e:
			sys.tracebacklimit = -1
			raise ImportError(f"Failed to import application [{name}], file [{l_path}], error: {e}.")
		"""
	return urlpatterns

# build url patterns.
def build_urls(views=[]):
	if dev0s.system.env.get("MIGRATIONS", format=bool, default=False):
		return []
	urls, landing_page = [], False
	for view in views:
		if view.type.lower() in ["view", "documentationview"] and not landing_page and view.landing_page:
			landing_page = True
			if view.auth_required:
				if view.except_xframe:
					urls.append(path("", view.__auth_xframe_except_view__))
					urls.append(path("/", view.__auth_xframe_except_view__))
				else:
					urls.append(path("", view.__auth_view__))
					urls.append(path("/", view.__auth_view__))
			else:
				if view.except_xframe:
					urls.append(path("", view.__xframe_except_view__))
					urls.append(path("/", view.__xframe_except_view__))
				else:
					urls.append(path("", view.__view__))
					urls.append(path("/", view.__view__))
		if view.auth_required:
			urls.append(
				path(gfp.clean(view.url, remove_first_slash=True, remove_last_slash=True)+"/", view.__auth_view__),
			)
		else:
			urls.append(
				path(gfp.clean(view.url, remove_first_slash=True, remove_last_slash=True)+"/", view.__view__),
			)
	return urls
	
# the django request object class.
class Request(Object):
	# do not forget the self.parameters's functions.
	def __init__(self,
		# the base path (required; if url path is null) [#1 argument].
		base=None,
		# the requests id (required) [#2 argument].
		id=None, 
		# the url path (optional).
		url=None,
		# the w3bsite.Website object (required).
		website=None,
		# authentication required.
		auth_required=False,
		# root permission required.
		root_required=False,
		# overwrite default maintenance (bool) (leave None to use website.maintenance).
		maintenance=None,
	):


		# docs.
		DOCS = {
			"module":"website.views.Request", 
			"initialized":False,
			"description":[], 
			"chapter": "Views", }

		# defaultss.
		Object.__init__(self)
		self.parameters = dev0s.response.parameters

		# variables.
		self.base = base
		self.id = id
		if url == None:
			self.url = gfp.clean(f"{base}/{id}/", remove_first_slash=True)
		else:
			self.url = gfp.clean(url.replace("/$id",f"/{self.id}"), remove_last_slash=True, remove_first_slash=True)+"/"
		if self.base == None: 
			self.base = gfp.clean(FilePath(self.url).base(back=1), remove_last_slash=True)+"/"
		self.url = gfp.clean(self.url, remove_first_slash=True).replace(".html", "")
		self.base = gfp.clean(self.base, remove_first_slash=True)
		self.type = "Request"
		self.website = website
		self.auth_required = auth_required
		self.root_required = root_required
		self.maintenance_ = maintenance
		if self.maintenance_ == None: self.maintenance_ = self.website.maintenance

		# checks.
		if self.website.__class__.__name__ not in ["Website"]:
			raise dev0s.exceptions.InvalidUsage(f"You forgot to pass the initialized w3bsite.Website object as the [website] parameter ({self.url}).")

	# response functions.
	def success(self, message, arguments={}):
		return dev0s.response.success(message, arguments, django=True)
	def error(self, error):
		return dev0s.response.error(error, django=True)
	def response(self, response):
		if isinstance(response, (JsonResponse, HttpResponse)):
			return response
		else:
			try:
				return JsonResponse(response.dict(safe=True), safe=False)
			except AttributeError:
				return JsonResponse(response, safe=False)

	# default responses.
	def _403(self, request=None):
		return self.error(f"Permission denied [403] ({self.url}).")
	def _404(self, request=None, error=None):
		return self.error(f"Page not found [404] ({self.url}).")
	def _500(self, request=None, error=None):
		info = utils.catch_error(error)
		return self.error(f"Server error [500] ({self.url}).")
	def _503(self, request=None):
		return self.error(f"Domain is under maintenance [503] ({self.url}).")
	def permission_denied(self, request=None):
		return self._403(request=request)
	def maintenance(self, request=None):
		return self._503(request=request)

	"""
	the view functions.
	 - the request() function must be overwritten.
	 - the __auth_view__ & __view__ are called before and should not be overwritten. """
	def request(self, request):
		return self.error(f"You did not define a default request() function for request [{self.url}].")
	def __view__(self, request, count=True):

		# count request.
		if count:
			self.website.metrics.count_api_request(request, data={
				"url":self.url,
			})

		# maintenance.
		if self.maintenance_:
			return self.maintenance(request=request)

		# make request.
		try: 
			response = self.request(request)
		except Exception as e: return self._500(request, error=e)
		return self.response(response)

		#
	def __auth_view__(self, request, count=True):
		
		# count request.
		if count:
			self.website.metrics.count_api_request(request, data={
				"url":self.url,
			})

		# check authenticated.
		if self.auth_required:
			response = self.website.users.authenticated(request)
			if not response.success: return self.response(response)

		# check root permissions.
		if self.root_required:
			response = self.website.users.root_permission(request)
			if not response.success: return self.response(response)

		# default __view__.
		return self.__view__(request, count=False)

	#

# the django view object class.
class View(Object):
	# do not forget the self.parameters's functions.
	def __init__(self, 
		# the base path (required; if url path is null) [#1 argument].
		base=None,
		# the views id (required) [#2 argument].
		id=None, 
		# the url path (optional).
		url=None,
		# the html path (optional).
		html=None,
		# the w3bsite.Website object (required).
		website=None,
		# enable if this view is the [/] landing page.
		landing_page=False,
		# authentication required.
		auth_required=False,
		# root permission required.
		root_required=False,
		# overwrite default maintenance (bool) (leave None to use website.maintenance).
		maintenance=None,
		# except xframe for the view.
		except_xframe=False,
		# the object type (do not edit).
		type="View",
	):

		# docs.
		DOCS = {
			"module":"website.views.View", 
			"initialized":False,
			"description":[], 
			"chapter": "Views", }

		# defaults.
		Object.__init__(self)
		self.parameters = dev0s.response.parameters

		# vars.
		self.base = base
		self.id = id
		self.website = website
		self.landing_page = landing_page
		self.type = type
		if url == None:
			self.url = gfp.clean(f"{base}/{id}/", remove_first_slash=True)
		else:
			self.url = gfp.clean(url.replace("/$id",f"/{self.id}"), remove_last_slash=True, remove_first_slash=True)+"/"
		if self.base == None: 
			self.base = gfp.clean(FilePath(self.url).base(back=1), remove_last_slash=True)+"/"
		if html != None:
			if len(html) >= len(f"/apps/") and html[:len(f"/apps/")] == f"/apps/": html = html[len(f"/apps"):]
			elif len(html) >= len(f"apps/") and html[:len(f"apps/")] == f"apps/": html = html[len(f"apps/"):]
			self.html = gfp.clean(html.replace("/$id",f"/{self.id}"))
			self.app = gfp.clean(self.html, remove_first_slash=True, remove_last_slash=True, remove_double_slash=True).split("/")[0]
		else:
			self.app = gfp.clean(self.base, remove_first_slash=True, remove_last_slash=True, remove_double_slash=True).split("/")[0]
			l_base = gfp.clean(str(self.base), remove_last_slash=True)+"/"
			if len(l_base) >= len(f"{self.app}/{self.app}") and l_base[:len(f"{self.app}/{self.app}")] == f"{self.app}/{self.app}": l_base = l_base[len(f"{self.app}/"):]
			elif len(l_base) >= len(f"{self.app}/") and l_base[:len(f"{self.app}/")] == f"{self.app}/": l_base = l_base[len(f"{self.app}/"):]
			self.html = gfp.clean(f"{self.app}/html/{l_base}/{self.id}.html")
		self.base = gfp.clean(self.base, remove_last_slash=True)+"/"
		self.url = gfp.clean(self.url.replace(".html", ""), remove_first_slash=True, remove_last_slash=True)+"/"
		self.html = gfp.clean(self.html)
		self.type = type
		self.auth_required = auth_required
		self.root_required = root_required
		self.maintenance_ = maintenance
		if self.maintenance_ == None: self.maintenance_ = self.website.maintenance
		self.except_xframe = except_xframe

		# check html.
		if self.type.lower() in ["view"] and "w3bsite/" not in self.html:
			path = FilePath(f"apps/{self.app}/")
			if Files.exists(path):
				path = Files.join(path, "html")
				if not Files.exists(path): 
					dev0s.response.log(f"&ORANGE&Creating&END& html base [{path}].")
					Files.create(path, directory=True)
				path = Files.join("apps", self.html)
				if not Files.exists(path):
					dev0s.response.log(f"&ORANGE&Creating&END& default html view [{path}].")
					base = FilePath(path).base(back=1)
					if not Files.exists(base): Files.create(base, directory=True)
					Files.copy(f"{SOURCE_PATH}/classes/django/lib/html/view.html", path)

		# checks.
		if self.website.__class__.__name__ not in ["Website"]:
			raise dev0s.exceptions.InvalidUsage(f"You forgot to pass the initialized w3bsite.Website object as the [website] parameter ({self.url}).")

		#

	# render functions.
	def render(self, 
		# the request (obj) (#1)
		request, 
		# overwrite default template data. #2
		template_data=None,
		# overwrite default html #3.
		html=None,
		# the response's status code.
		status=200,
	):
		if html == None: html = self.html
		return render(request, str(html), self.template(template_data), status=status)
	def error(self, 
		# the django request parameter.
		request, 
		# the error title.
		title="Warning!",
		# the error title.
		message="Some error occured (#000000).",
		# the error icon (the static directory is root).
		icon="media/icons/warning.png", 
		# the redirect button text (right button).
		redirect_button="Ok",
		# the redirect url.
		redirect="/dashboard/home/",
		# overwrite default template data.
		template_data=None,
		# pass arguments by dict.
		serialized={},
	):
		if serialized != {}:
			title, message, icon, redirect_button, redirect = Dictionary(serialized).unpack({
				"title":"Warning!",
				"message":"Some error occured (#000000).",
				"icon":"media/icons/warning.png",
				"redirect_button":"Ok",
				"redirect":"/dashboard/home/",
			})
		if template_data == None: template_data = self.website.template_data
		if template_data == None: template_data = {}
		title = title.replace("\n","")
		message = message.replace("\n","")
		template_data["ERROR"] = {
			"title":title,
			"message":message,
			"icon":icon,
			"redirect":redirect,
			"redirect_button":redirect_button,
		}
		return self.render(request=request, template_data=template_data, html=f"w3bsite/classes/apps/defaults/html/error.html")

	# default renders.
	def _403(self, request, template_data=None):
		return self.render(request=request, template_data=template_data, status=403, html=f"w3bsite/classes/apps/defaults/html/permission_denied.html")
	def _404(self, request, template_data=None):
		return self.render(request=request, template_data=template_data, status=404, html=f"w3bsite/classes/apps/defaults/html/404.html")
	def _500(self, request, template_data=None, error=None):
		if template_data == None: template_data = self.website.template_data
		if template_data == None: template_data = {}
		info = utils.catch_error(error)
		traceback, debug = None, False
		if not dev0s.env.get("PRODUCTION", format=bool, default=True) or dev0s.env.get("DEBUG", format=bool, default=False):
			debug = True
			traceback = info["traceback"]
		return self.render(request=request, html=f"w3bsite/classes/apps/defaults/html/500.html", status=500, template_data=self.template(old=template_data, new={
			"ERROR_ID":str(info["id"]),
			"DEBUG":str(debug),
			"TRACEBACK":str(traceback),
			"URL":self.url,
		}))
	def _503(self, request, template_data=None):
		return self.render(request=request, template_data=template_data, status=503, html=f"w3bsite/classes/apps/defaults/html/maintenance.html")
	def permission_denied(self, request, template_data=None):
		return self._403(request, template_data=template_data)
	def maintenance(self, request, template_data=None):
		return self._503(request, template_data=template_data)

	# append template data.
	def template(self, new={}, old=None, safe=False):
		if old == None: old = self.website.template_data
		return utils.template(old=old, new=new, safe=safe)

	# the view functions.
	# the view() function can be overwritten.
	# the __auth_view__ & __view__ are called before and should not be overwritten.
	def view(self, request):
		return self.render(request)
	def __view__(self, request, count=True):

		# count request.
		if count:
			self.website.metrics.count_web_request(request, data={
				"url":self.url,
			})

		# maintenance.
		if self.maintenance_:
			return self.maintenance(request=request)

		# default view.
		try: 
			return self.view(request)
		except Exception as e: return self._500(request, error=e)
		return self.render(request=request)

		#
	@method_decorator(login_required)
	def __auth_view__(self, request, count=True):

		# count request.
		if count:
			self.website.metrics.count_web_request(request, data={
				"url":self.url,
			})

		# check root permission.
		if self.root_required:
			response = self.website.users.root_permission(request)
			if not response.success: return self._404(request)

		# return __view__.
		return self.__view__(request, count=False)

		#
	@method_decorator(xframe_options_exempt)
	def __xframe_except_view__(self, request):

		# count request.
		if count:
			self.website.metrics.count_web_request(request, data={
				"url":self.url,
			})

		# return __view__.
		return self.__view__(request, count=False)
	@method_decorator(login_required)
	@method_decorator(xframe_options_exempt)
	def __auth_xframe_except_view__(self, request):

		# count request.
		if count:
			self.website.metrics.count_web_request(request, data={
				"url":self.url,
			})

		# check root permission.
		if self.root_required:
			response = self.website.users.root_permission(request)
			if not response.success: return self._404(request)

		# return __view__.
		return self.__view__(request, count=False)
	#

#

















# template data object !!! DEPRICATED !!!.
class TemplateData(Object):
	def __init__(self, data={}):
		# defaults.
		Object.__init__(self)
		self.assign(data)
	# return raw data.
	def raw(self):
		return self.attributes()
		

#


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from w3bsite.classes.config import *
from w3bsite.classes import security, heroku, namecheap, utils, git, stripe, rate_limit, deployment, vps, defaults, apps, views, electron
from w3bsite.classes import database as _database_
from w3bsite.classes.apps.metrics import metrics
from w3bsite.classes.apps.logging import logging
import django as pypi_django

# the main website class.
class Website(dev0s.cli.CLI,Traceback):
	def __init__(self,
		#
		# General.
		# 	the root path.
		root=None, # example: FilePath(__file__).base(back=1).replace("./","")
		# 	the root domain.
		domain=None,
		# 	the website name.
		name=None,
		# 	the database path (optional).
		database=None,
		# 	the library path (optional).
		library=None,
		#	the website's description.
		description="",
		#a
		# Deployment.
		# 	remote depoyment, options: [local, vps, heroku].
		remote="local",
		#
		# Developers.
		#	the developer users (emails).
		developers=[],
		#
		# Django.
		#	maintenance boolean.
		maintenance=False,
		# 	the template data (only required when running the website) (overwrites the w3bsite template data keys).
		template_data={},
		#	2fa required for login.
		_2fa=False,
		#
		# Organization.
		# 	the author's / comitters name.
		author=None,
		# 	the admin's email.
		email=None,
		# 	the organization name.
		organization=None,
		# 	the organization unit.
		organization_unit="Information Technology",
		# 	the organization country code.
		country_code="NL",
		# 	the organization localization's city name.
		city=None,
		# 	the organization localization's province / state.
		province=None,
		#
		# AES.
		#	the passphrase of the aes master-key (defaults is no passphrase).
		aes_passphrase=None,
		#
		# Namecheap.
		#	namecheap enabled.
		namecheap_enabled=True,
		# 	your namecheap username.
		namecheap_username=None,
		# 	your namecheap api key.
		namecheap_api_key=None,
		#
		# Firebase.
		#	firebase enabled.
		firebase_enabled=True,
		# 	your firebase admin service account key, (dict) [https://console.firebase.google.com > Service Accounts > Firebase admin].
		firebase_admin={},
		# 	your firebase sdk javascript configuration, (dict) [https://console.firebase.google.com > Settings > General > Web JS SDK].
		firebase_js={},
		#
		# Stripe.
		# enable strip.e
		stripe_enabled=True,
		# 	your stripe secret key (str) [https://stripe.com > Dashboard > Developer > API Keys > Secret Key].
		stripe_secret_key=None,
		# 	your stripe publishable key (str) [https://stripe.com > Dashboard > Developer > API Keys > Secret Key].
		stripe_publishable_key=None,
		#	the stripe subscriptions.
		#		do not edit the plan & product names after creation.
		#		price changes are not supported yet, will be in the future.
		stripe_subscriptions={
			#"vserver": {
			#	"basic": {
			#		"rank":1,
			#		"price":50,
			#		"currency":"eur",
			#		"favicon":"https://raw.githubusercontent.com/vandenberghinc/public-storage/master/vserver/icon/icon.png"
			#	},
			#	"premium": {
			#		"rank":2,
			#		"price":100,
			#		"currency":"eur",
			#	},
			#	"pro": {
			#		"rank":3,
			#		"price":250,
			#		"currency":"eur",
			#	},
			#}
		},
		# 	the stripe products.
		stripe_products={
			#"vserver": {
			#	"basic": {
			#		"rank":1,
			#		"price":50,
			#		"currency":"eur",
			#		"favicon":"https://raw.githubusercontent.com/vandenberghinc/public-storage/master/vserver/icon/icon.png"
			#	},
			#	"premium": {
			#		"rank":2,
			#		"price":100,
			#		"currency":"eur",
			#	},
			#	"pro": {
			#		"rank":3,
			#		"price":250,
			#		"currency":"eur",
			#	},
			#}
		},
		#
		#
		# Smtp Email
		# 	sending emails [https://gmail.com > Security > Enable Unsafe 3th party applications / generate an app password (advised)].
		email_enabled=True,
		email_address=None,
		email_password=None,
		email_smtp_host="smtp.gmail.com",
		email_smtp_port=587,
		#
		# VPS.
		# 	the public ip of the vps.
		vps_ip=None,
		#	the vps ssh port.
		vps_port=22,
		#	the executing username on the vps.
		vps_username=None,
		#
		# Additional Options.
		# 	prevent heroku deployment.
		prevent_heroku_deployment=False,
		# 	purchase tls/ssl certificate from namecheap.
		purchase_tls_certificate=False,
		# 	interactive mode.
		interactive=False,
		#	production mode.
		production=True,
		# 	debug mode.
		debug=False,
		# 	the users sub path.
		users_subpath="users/",
		# 	id users by username.
		id_by_username=True,
		#
		# the logs.
		log_level=dev0s.defaults.options.log_level,
		# styling options.
		styling={},
		#
		# 
		# optionally initizialize from a serialized dict (for config.py) (still requires parameters: root).
		serialized=None,
	):	
		# docs.
		DOCS = {
			"module":"w3bsite.Website", 
			"initialized":False,
			"description":[], 
			"chapter": "Website", }

		# defaults.
		Traceback.__init__(self, traceback="w3bsite.Website",)

		# w3bsite values.
		self.SOURCE_PATH = SOURCE_PATH

		# argurments.
		if root != None:
			if root[len(root)-1] != "/": root += "/"
			if root[-3:] == "/./": root = root[:-3]
			if root[len(root)-1] != "/": root += "/"
			root = root.replace("//","/").replace("//","/")
			if firebase_admin != None:
				if isinstance(firebase_admin, str) and firebase_admin[:len("__defaults__/env")] == "__defaults__/env":
					firebase_admin = f"{root}/{firebase_admin}".replace("//","/").replace("//","/")
		self.root = gfp.clean(root)
		self.name = name
		self.description = description
		self.author = author
		self.email = email
		self.city = city
		self.province = province
		self.organization = organization
		self.organization_unit = organization_unit
		self.country_code = country_code
		self.domain = domain
		self.namecheap_username = namecheap_username
		self.namecheap_api_key = namecheap_api_key
		self.firebase_admin = firebase_admin
		self.firebase_js = firebase_js
		self.stripe_secret_key = stripe_secret_key
		self.stripe_publishable_key = stripe_publishable_key
		self.email_address = email_address
		self.email_password = email_password
		self.email_smtp_host = email_smtp_host
		self.email_smtp_port = email_smtp_port
		self.purchase_tls_certificate = purchase_tls_certificate
		self.aes_passphrase = aes_passphrase
		self.stripe_subscriptions = stripe_subscriptions
		self.stripe_products = stripe_products
		self.developers = developers
		if remote == None: self.remote = remote
		else: self.remote = remote.lower()
		self.vps_ip = vps_ip
		self.vps_port = vps_port
		self.vps_username = vps_username
		self._2fa = _2fa
		self.template_data = template_data
		self.database = database
		self.library = library
		self.prevent_heroku_deployment = prevent_heroku_deployment
		self.interactive = interactive
		self.production = production
		self.debug = debug
		self.maintenance = maintenance
		self.firebase_enabled = firebase_enabled
		self.stripe_enabled = stripe_enabled
		self.namecheap_enabled = namecheap_enabled
		self.email_enabled = email_enabled
		self.log_level = log_level
		self.users_subpath = users_subpath
		self.id_by_username = id_by_username
		self.styling = styling
		# checks.
		if self.database == None: self.database = f"/etc/{self.domain}/"
		if self.library == None: self.library = f"/usr/local/lib/{self.domain}/"
		if self.domain != None: self.domain = utils.naked_url(self.domain)

		# serialize.
		if serialized == None or serialized == {}:
			self.serialize(save=True)
		else:
			if isinstance(serialized, str):
				serialized = Files.load(serialized, format="json")
			self.init_from_serialized(serialized=serialized)
		#if root != dev0s.utils.__execute_script__("pwd").replace("\n",""):
		#	raise ValueError("The ")

		# init.
		response = self.initialize()
		if not response.success: response.crash()

		#

	# initialize.
	def initialize(self):

		# overall arguments.
		response = dev0s.response.parameters.check(
			traceback=self.__traceback__(),
			parameters={
				"root":self.root,
				"database":self.database,
				"library":self.library,
				"name":self.name,
				"organization":self.organization,
				"domain":self.domain,
				"developers":self.developers,
				"_2fa":self._2fa,
			})
		if not response.success: return response

		# namecheap arguments.
		if self.namecheap_enabled:
			response = dev0s.response.parameters.check(
				traceback=self.__traceback__(),
				parameters={
					"author":self.author,
					"email":self.email,
					"city":self.city,
					"province":self.province,
					"organization_unit":self.organization_unit,
					"country_code":self.country_code,
					"namecheap_username":self.namecheap_username,
					"namecheap_api_key":self.namecheap_api_key,
				})
			if not response.success: return response

		# firebase arguments.
		if self.firebase_enabled:
			response = dev0s.response.parameters.check(
				traceback=self.__traceback__(),
				parameters={
					"firebase_admin":self.firebase_admin,
					"firebase_js":self.firebase_js,
				})
			if not response.success: return response

		# stripe arguments.
		if self.stripe_enabled:
			response = dev0s.response.parameters.check(
				traceback=self.__traceback__(),
				parameters={
					"stripe_secret_key":self.stripe_secret_key,
					"stripe_publishable_key":self.stripe_publishable_key,
					"stripe_subscriptions":self.stripe_subscriptions,
					"stripe_products":self.stripe_products,
				})
			if not response.success: return response

		# email arguments.
		if self.email_enabled:
			response = dev0s.response.parameters.check(
				traceback=self.__traceback__(),
				parameters={
					"email_address":self.email_address,
					"email_password":self.email_password,
					"email_smtp_host":self.email_smtp_host,
					"email_smtp_port":self.email_smtp_port,
				})
			if not response.success: return response

		# remote: vps arguments.
		if self.remote in ["vps"]:
			response = dev0s.response.parameters.check(
				traceback=self.__traceback__(),
				parameters={
					"vps_ip":self.vps_ip,
					"vps_port":self.vps_port,
					"vps_username":self.vps_username,
				})
			if not response.success: return response

		# create database.
		if not Files.exists(self.database):
			dev0s.response.log(f"&ORANGE&Root permission&END& required to create database [{self.database}].")
			Files.create(f"{self.database}", sudo=True, directory=True)
			Files.chmod(path=self.database, permission=700, sudo=True)
			Files.chown(path=self.database, owner=dev0s.defaults.vars.user, group=dev0s.defaults.vars.group, sudo=True)
		for dir in [
			Files.join(self.database, "keys"),
		]:
			if not Files.exists(dir): Files.create(dir, directory=True)

		# set variables.
		self.live = utils.equalize_path(self.root, striplast=True) == utils.equalize_path(self.library, striplast=True)
		self.http_domain = f"http://{self.domain}"
		self.https_domain = f"https://{self.domain}"

		# environment for settings.py.
		os.chdir(gfp.clean(self.root))
		if not os.path.exists("__defaults__/env"): os.mkdir("__defaults__/env")
		SECRET_KEY = dev0s.env.get("DJANGO_SECRET_KEY", default=None)
		if SECRET_KEY == None:  SECRET_KEY = String().generate(length=128, capitalize=True, digits=True, special=True)
		dev0s.env.export(export="__defaults__/env/json", env={
			"DJANGO_SECRET_KEY":SECRET_KEY,
			"WEBSITE_BASE":gfp.base(SOURCE_PATH),
			"DOMAIN":str(self.domain),
			"DATABASE":str(self.database),
			"PRODUCTION":str(self.production),
			"DEBUG":self.debug,
		})

		# template data.
		colors = {
			"white":"#FAFAFA",
			"light_white":"#E9F0FD",
			"grey":"#E5E5E5",
			"light_grey":"#D6D6D6",
			"dark_grey":"#424242",
			"blue":"#5A8FE6",
			"purple":"#323B83",#"#B32FCA",
			#"purple":"#9B00AA",
			"red":"#FD304E",
			"pink":"#F62B7D",
			"orange":"#FF8800",
			"green":"#006633",
			"darkest":"#1F2227",
			"darker": "#20242A",
			"dark": "#262B30",
			# background color.
			"topbar":"#FAFAFA",#1F2227", #"#FAFAFA",
			"background":"#323B83",#"#E7E9EF", #"#FAFAFA",
			"topbar_darkmode":"#1F2227",#1F2227", #"#FAFAFA",
			"background_darkmode":"#1F2227",#"#E7E9EF", #"#FAFAFA",
			"background_img":None,
			# elements.
			"widgets":"#FAFAFA",
			"widgets_reversed":"#323B83",#"#1F2227",
			"widgets_darkmode":"#20242A",
			"widgets_reversed_darkmode":"#323B83",#"#1F2227",
			# text.
			"text":"#1F2227",
			"text_reversed":"#FAFAFA",
			"text_darkmode":"#FAFAFA",
			"text_reversed_darkmode":"#FAFAFA",
			# input & textareas.
			"input_txt":"#6C6B6D",
			"input_txt_reversed":"#FAFAFA",
			"input_bg":"#E9F0FD", #"#FAFAFA", 
			"input_bg_reversed":"#323B83",
			
			# buttons.
			"button_txt":"#FAFAFA",
			"button_txt_reversed":"#1F2227",
			"button_bg":"#323B83",
			"button_bg_reversed":"#FAFAFA",
			# custom colors.
			# ...
		}

		# check styling.
		d = {}
		for key,value in self.styling.items(): d[key.upper()] = value
		styling = Dictionary(d).check(default={
			# styling options.
			"LEFTBAR_WIDTH":"280px",
			"RIGHTBAR_WIDTH":"280px",
			"TOPBAR_HEIGHT":"50px" ,
			})
		self.template_data = Dictionary(utils.__append_dict__(new=self.template_data, overwrite=True,  old=styling))

		# add safe template data.
		self.template_data = Dictionary(utils.__append_dict__(new=self.template_data, overwrite=True,  old={
			# colors.
			"COLORS":colors,
			# wbsite info.
			"NAME":self.name,
			"DOMAIN":self.domain,
			"PRODUCTION":self.production,
			"AUTHOR":self.author,
			"ORGANIZATION":self.organization,
			# include other tempalte data's.
			"STRIPE":{},
			"FIREBASE":{},
			# options.
			"2FA":self._2fa,
			# styling options.
			"LEFTBAR_WIDTH":280, # px
			"RIGHTBAR_WIDTH":280, # px
			"TOPBAR_HEIGHT":50 #px
		}))

		# defaults objects.
		if not Files.exists(Files.join(self.root, ".version")): Files.create(Files.join(self.root, ".version"), data="1.0.00")
		self.version = Version(Files.load(Files.join(self.root, ".version")))
		self.aes = dev0s.encryption.AsymmetricAES(
			private_key=f"{self.database}/keys/{ALIAS}/private_key",
			public_key=f"{self.database}/keys/{ALIAS}/public_key",
			passphrase=self.aes_passphrase)
		if  not Files.exists(self.aes.rsa.public_key):
			response = self.aes.generate_keys()
			if not response.success:
				return response
		response = self.aes.load_keys()
		if not response.success: return response
		self.logging = logging.Logging(attributes={
			"database":self.database,
			"live":self.live
		})
		self.defaults = defaults.Defaults({
			"root":self.root,
			"library":self.library,
			"database":self.database,
			"description":self.description,
			"name":self.name,
			"author":self.author,
			"email":self.email,
			"organization":self.organization,
			"country_code":self.country_code,
			"province":self.province,
			"city":self.city,
			"organization_unit":self.organization_unit,
			"domain":self.domain,
			"https_domain":self.https_domain,
			"developers":self.developers,
			"remote":self.remote,
			"live":self.live,
			"interactive":self.interactive,
			"_2fa":self._2fa,
			"maintenance":self.maintenance,
			"users_subpath":self.users_subpath,
			"template_data":self.template_data,
			"id_by_username":self.id_by_username,
			"aes":self.aes, 
			"logging":self.logging,
			"version":self.version,})
		self.template = self.defaults.template
		self.utils = utils.Utils(attributes=self.defaults.attributes())

		# objects.
		self.metrics = metrics.Metrics(database=self.database)
		self.security = security.Security(
			defaults=self.defaults,)
		if self.firebase_enabled and dev0s.defaults.vars.os not in ["macos"]:
			from w3bsite.classes import firebase
			self.firebase = firebase.Firebase(
				key=self.firebase_admin,
				firebase_js=self.firebase_js,
				defaults=self.defaults)
			self.firestore = self.firebase.firestore
			self.db = _database_.Database(
				firestore=self.firebase.firestore,
				path=self.database,
				live=self.live, )
		else:
			self.firebase = None
			self.firestore = None
			self.db = _database_.Database(path=self.database, live=self.live)
		if self.stripe_enabled:
			self.stripe = stripe.Stripe(
				secret_key=self.stripe_secret_key,
				publishable_key=self.stripe_publishable_key,
				subscriptions=self.stripe_subscriptions,
				products=self.stripe_products,
				defaults=self.defaults, )
		else: self.stripe = None
		self.rate_limit = rate_limit.RateLimit(
			db=self.db,
			defaults=self.defaults,)
		if not os.path.exists("__defaults__/django/settings.py"): raise ImportError(f"Invalid website hierarchy, unable to find: __defaults__.django.settings, required location: {self.root}/__defaults__/django/settings.py")
		os.environ.setdefault('DJANGO_SETTINGS_MODULE', f'__defaults__.django.settings')
		dev0s.system.env.import_(env="__defaults__/env/json")
		#dev0s.response.log("&RED&IMPORTING SETTINGS&END&")
		from __defaults__.django import settings
		#dev0s.response.log("&RED&ATTEMPT DJANGO SETUP&END&")
		pypi_django.setup()
		#try:pypi_django.setup()
		#except: a=1
		#dev0s.response.log("&GREEN&DONE&END&")
		from w3bsite.classes import django, users
		self.django = django.Django(
			security=self.security,
			defaults=self.defaults,)
		self.users = users.Users(
			email_address=self.email_address,
			email_password=self.email_password,
			smtp_host=self.email_smtp_host,
			smtp_port=self.email_smtp_port,
			firestore=self.firestore,
			email_enabled=self.email_enabled,
			db=self.db,
			stripe=self.stripe,
			django=self.django,
			defaults=self.defaults,)
		if self.namecheap_enabled:
			self.namecheap = namecheap.Namecheap(
				username=self.namecheap_username,
				api_key=self.namecheap_api_key,
				root=self.root,
				domain=self.domain,
				email=self.email_address,)
		else: self.namecheap = None
		self.git = git.Git(
			defaults=self.defaults,)#
		# mode dependend objects.
		self.deployment = None
		self.heroku = None
		self.vps = None
		if self.remote in ["local", "vps"]:
			self.deployment = deployment.Deployment(
				root=self.root,
				library=self.library,
				name=self.name,
				domain=self.domain,
				database=self.database,
				remote=self.remote,
				vps_ip=self.vps_ip,
				vps_username=self.vps_username,
				namecheap=self.namecheap,)
			if self.remote in ["vps"]:
				self.vps = vps.VPS(
					ip=self.vps_ip,
					port=self.vps_port,
					username=self.vps_username,
					namecheap=self.namecheap,
					deployment=self.deployment,
					defaults=self.defaults,)
		elif self.remote == "heroku":
			self.heroku = heroku.Heroku(
				root=self.root,
				domain=self.doman,
				name=self.name,
				namecheap=self.namecheap,)
		elif self.remote != None:
			raise dev0s.response.error(f"Selected an invalid remote [{self.remote}], options: [local, vps, heroku]")
		self.electron = electron.Electron(attributes=self.defaults.attributes())
		self.defaults.website = self
		self.apps = apps.Apps(
			template_data=self.template_data,
			rate_limit=self.rate_limit,
			users=self.users,
			stripe=self.stripe,
			utils=self.utils,
			defaults=self.defaults,)
		self.metrics = metrics.Metrics(database=self.database)

		# defaults.
		dev0s.cli.CLI.__init__(self,
			modes={
				"Development":"*chapter*",
				"    --create":"Create the website.",
				"    --create-app dashboard":"Create a new application.",
				"    --start --developer":"Start the production webserver.",
				"Production":"*chapter*",
				"    --start":"Start the nginx webserver.",
				"    --stop":"Stop the nginx webserver.",
				"    --restart":"Restart the nginx webserver.",
				"    --status":"Retrieve the status of the nginx webserver.",
				"Deployment":"*chapter*",
				"    --generate-aes-passphrase":"Generate an aes passphrase.",
				"    --generate-tls":"Generate a tls certificate.",
				"    --activate-tls":"Activate the generated tls certificate.",
				"    --bundle-tls /path/to/downloaded/cert/":"Bundle the downloaded signed certificate from the CA.",
				"    --deploy":"Deploy the website.",
				"Electron":"*chapter*",
				"    --electron":"Electron app.",
				"        --build":"Build the electron app.",
				"        --start":"Start the electron app for development purposes.",
				"        --deploy":"Deploy the django website into an electron app.",
				"Logs":"*chapter*",
				"    --tail [optional: --nginx]":"Tail the (deployed) website.",
				"Documentation":"*chapter*",
				"    -h / --help":"Show the documentation.",
			},
			options={
				#"-c":"Do not clear the logs.",
			},
			alias=ALIAS,
			executable=__file__,
		)

		# logs.
		if dev0s.defaults.options.log_level >= 2:
			print(f"Website: {self.name}")
			print(f" * domain: {self.domain}")
			print(f" * root: {self.root}")
			print(f" * library: {self.library}")
			print(f" * database: {self.database}")
			print(f" * pwd: {dev0s.defaults.pwd()}")
			print(f" * remote: {self.remote}")
			print(f" * live: {self.live}")

		# handler.
		return dev0s.response.success(f"Successfully initialized website [{self.name}].")

		#

	# cli.
	def cli(self):

		# check args.
		self.arguments.check(exceptions=["--log-level", "--non-interactive", "--create-alias", "--developer", "--nginx", "--code-update", "--forced", "-f", "--w3bsite"])

		# activate enc.
		if self.remote in ["vps"] and not self.vps.live:
			if not ssht00ls_agent.generated:
				response = ssht00ls_agent.generate()
				if not response.success: self.stop(response=response, json=dev0s.defaults.options.json)
			elif not ssht00ls_agent.activated:
				response = ssht00ls_agent.activate()
				if not response.success: self.stop(response=response, json=dev0s.defaults.options.json)

		# ______________________________________________________________________________________
		#
		# Defaults.
		#

		# help.
		if self.arguments.present("-h"):
			self.docs(success=True)

		# version.
		elif self.arguments.present(['--version']):
			self.stop(message=f"{ALIAS} version:"+Files.load(f"{SOURCE_PATH}/.version").replace("\n",""), json=dev0s.defaults.options.json)

		# ______________________________________________________________________________________
		#
		# Electron
		#

		# electron.
		elif self.arguments.present("--electron"):

			# build electron source dode.
			if self.arguments.present("--build"):
				response = self.electron.build(
					build=self.arguments.get("--build", required=False, default=None),
				)
				dev0s.response.log(response=response)

			# developer start the electron app.
			elif self.arguments.present("--start"):
				response = self.electron.start()
				dev0s.response.log(response=response)

			# deploy electron.
			elif self.arguments.present("--deploy"):
				response = self.electron.deploy()
				dev0s.response.log(response=response)

			# invalid.
			else: self.invalid(chapter="electron")

		# ______________________________________________________________________________________
		#
		# Development.
		#

		# developer start.
		elif self.arguments.present('--start') and self.arguments.present('--developer'):
			default_host, default_port = "127.0.0.1", 8000
			if ":" in self.domain:
				try: 
					default_host, default_port = self.domain.split(":")
				except: a=1
			response = self.django.start(
				host=self.arguments.get("--host", required=False, default=default_host), 
				port=self.arguments.get("--port", required=False, default=default_port),
			)
			self.stop(response=response)

		# create.
		elif self.arguments.present("--create"):
			response = self.create()
			dev0s.response.log(response=response)

		# create app.
		elif self.arguments.present("--create-app"):
			response = self.django.create_app(name=self.arguments.get("--create-app"))
			dev0s.response.log(response=response)

		# ______________________________________________________________________________________
		#
		# Live.
		#

		# deploy.
		elif self.arguments.present('--deploy'):
			self.stop(response=self.deploy(
				code_update=self.arguments.present("--code-update"),
				reinstall=self.arguments.present("--reinstall"),
			))

		# start.
		elif self.arguments.present('--start'):
			response = self.deployment.start()
			self.stop(response=response)

		# stop.
		elif self.arguments.present('--stop'):
			response = self.deployment.stop()
			self.stop(response=response)

		# restart.
		elif self.arguments.present('--restart'):
			response = self.deployment.restart()
			self.stop(response=response)

		# status.
		elif self.arguments.present('--status'):
			response = self.deployment.status()
			if not response.success:
				self.stop(response=response)
			else:
				print(response.status)

		# reset logs.
		elif self.arguments.present('--reset-logs'):
			response = self.deployment.reset_logs()
			self.stop(response=response)

		# tail.
		elif self.arguments.present('--tail'):
			response = self.deployment.tail(
				nginx=self.arguments.present("--nginx"),
				debug=self.arguments.present("--debug"),
			)
			if not response.success:
				self.stop(response=response)
			else:
				print(response.logs)

		# ______________________________________________________________________________________
		#
		# TLS & AES
		#

		# generate tls.
		elif self.arguments.present("--generate-tls"):
			response = self.deployment.generate_tls()
			dev0s.response.log(response=response)

		# activate tls.
		elif self.arguments.present("--activate-tls"):
			response = self.deployment.activate_tls()
			dev0s.response.log(response=response)

		# bundle tls.
		elif self.arguments.present("--bundle-tls"):
			response = self.deployment.bundle_tls(directory=self.arguments.get("--bundle-tls"))
			dev0s.response.log(response=response)

		# generate aes passphrase.
		elif self.arguments.present("--generate-aes-passphrase"):
			print(f"Generated AES Passphrase: {utils.__generate__(length=64, capitalize=True, digits=True)}")

		# invalid.
		else: self.invalid()

		#

	# management.
	def deploy(self, code_update=False, reinstall=False, log_level=0):

		# check git.
		loader = dev0s.console.Loader("Checking git repository ...")
		response = self.git.installed()
		if not response.success: 
			loader.stop(success=False)
			return response
		elif not response["installed"]:
			loader.stop()
			loader = dev0s.console.Loader("Installing git repository ...")
			response = self.git.install()
			if not response.success: 
				loader.stop(success=False)
				return response
			loader.stop()
		else: loader.stop()

		# heroku remote.
		if self.remote in ["heroku"]:

			# deploy on heroku.
			response = self.heroku.deploy(log_level=log_level)
			if not response.success: return response

		# local remote (& local when vps live).
		elif (self.remote in ["local"]) or (self.remote in ["vps"] and self.live):

			# deploy on local.
			response = self.deployment.deploy(code_update=code_update, reinstall=reinstall, log_level=log_level)
			if not response.success: return response

		# vps remote.
		elif self.remote in ["vps"]:

			# push to vps.
			response = self.vps.deploy(code_update=code_update, reinstall=reinstall, log_level=log_level)
			if not response.success: return response


		# check dns settings.
		if not code_update:
			response = self.check_dns(log_level=log_level)
			if response.error != None: return response
			return dev0s.response.success(f"Successfully deployed website https://{self.domain}.")
		else:
			return dev0s.response.success(f"Successfully deployed the code updates of website https://{self.domain}.")

		#
	def check_dns(self, log_level=0):
		
		# remote: heroku.
		if self.remote in ["heroku"]:
			response = self.heroku.check_dns(log_level=log_level)
			if not response.success: return response

		# remote: local & vps.
		elif self.remote in ["local", "vps"]:
			if not (self.remote in ["vps"] and self.live):
				response = self.deployment.check_dns(log_level=log_level)
				if not response.success: return response

		# handlers.
		return dev0s.response.success(f"Successfully checked the dns settings of website [{self.name}].", log_level=log_level)

		#
	def create(self):

		# create root.
		if not Files.exists(self.root):
			#return dev0s.response.error(f"Website [{self.root}] already exists.")
			os.mkdir(self.root)

		# create django.
		loader = dev0s.console.Loader("Creating website ...")
		response = self.django.create()
		loader.stop(success=response["success"])
		if response.error != None: return response

		# generate tls.
		loader = dev0s.console.Loader("Generating tls ...")
		response = self.security.generate_tls()
		loader.stop(success=response["success"])
		if response.error != None: return response

		# handlers.
		return dev0s.response.success(f"Successfully created website [{self.name}].")

		#

	# serialization.
	def serialize(self, save=False):
		# serizalized is used by the config.py in deployment.
		# all non secret variables are stored in __defaults__/env/website inside the root/. directory .
		# all secret variables are stored in environment variables from env.sh & env.json.
		# all keys of the stored secrets are stored in the serialzed __defaults__/env/website.

		# serialize and save all non secret variables in the __defaults__/env/website.
		serialized = {
			"root":self.root,
			"database":self.database,
			"library":self.library,
			"description":self.description,
			"name":self.name,
			"email":self.email,
			"city":self.city,
			"province":self.province,
			"organization":self.organization,
			"organization_unit":self.organization_unit,
			"country_code":self.country_code,
			"domain":self.domain,
			"author":self.author,
			#"namecheap_username":self.namecheap_username,
			#"namecheap_api_key":self.namecheap_api_key,
			#"firebase_admin":self.firebase_admin,
			#"stripe_secret_key":self.stripe_secret_key,
			#"stripe_publishable_key":self.stripe_publishable_key,
			#"email_address":self.email_address,
			#"email_password":self.email_password,
			"email_smtp_host":self.email_smtp_host,
			"email_smtp_port":self.email_smtp_port,
			"purchase_tls_certificate":self.purchase_tls_certificate,
			"stripe_subscriptions":self.stripe_subscriptions,
			"stripe_products":self.stripe_products,
			"developers":self.developers,
			"remote":self.remote,
			"vps_ip":self.vps_ip,
			"vps_port":self.vps_port,
			"vps_username":self.vps_username,
			"_2fa":self._2fa,
			"firebase_enabled":self.firebase_enabled,
			"stripe_enabled":self.stripe_enabled,
			"namecheap_enabled":self.namecheap_enabled,
			"email_enabled":self.email_enabled,
			"log_level":self.log_level,
			"users_subpath":self.users_subpath,
			"id_by_username":self.id_by_username,
			"styling":self.styling,
			"debug":self.debug,
		}

		# save all secret variables in secrets.
		local_security = security.Security(
			# the root path.
			root=self.root,)
		def __handle_dict__(base="", dictionary={
			"FIREBASE_ADMIN":self.firebase_admin,
			"FIREBASE_JS":self.firebase_js,
			"STRIPE": {
				"SECRET_KEY":self.stripe_secret_key,
				"PUBLISHABLE_KEY":self.stripe_publishable_key,
			},
			"EMAIL": {
				"ADDRESS":self.email_address,
				"PASSWORD":self.email_password,
			},
			"AES_PASSPHRASE":self.aes_passphrase,
			"NAMECHEAP_USERNAME":self.namecheap_username,
			"NAMECHEAP_API_KEY":self.namecheap_api_key,
			# add these for the settings.py
			"DOMAIN":self.domain,
			"DATABASE":self.database,
			"PRODUCTION":self.production,
			"MAINTENANCE":self.maintenance,
			"LOG_LEVEL":self.log_level,
		}):
			stored = []
			for key, value in dictionary.items():
				key = key.upper().replace("-","_").replace(".","_")
				if base == "": full_key = key
				else: full_key = base+"_"+key
				_format_ = None
				if isinstance(value, str):
					_format_ = "str"
					local_security.set_secret_env(full_key, value)
				elif isinstance(value, int):
					_format_ = "int"
					local_security.set_secret_env(full_key, str(value))
				elif isinstance(value, bool):
					_format_ = "bool"
					local_security.set_secret_env(full_key, str(value))
				elif value == None:
					_format_ = "null"
					local_security.set_secret_env(full_key, "None")
				elif isinstance(value, dict):
					_format_ = "dict"
					stored += __handle_dict__(dictionary=value, base=full_key)
				else:
					raise ValueError(f"Cannot secretly serialize [{key}:{value}].")
				#print(f"Stored secret environment variable: {key}.")
				stored.append([full_key, _format_])
			return stored
		__handle_dict__()

		# save.
		if save: 
			Files.save(f"{self.root}/__defaults__/env/website", serialized, format="json")


		# return serizaled.
		return serialized

		#
	def init_from_serialized(self, serialized=None):

		# load non secrets.
		if serialized == None:
			serialized = Files.load(f"{self.root}/__defaults__/env/website", format="json")
		#self.root = serialized["root"]
		self.database = serialized["database"]
		self.library = serialized["library"]
		self.description = serialized["description"]
		self.name = serialized["name"]
		self.author = serialized["author"]
		self.email = serialized["email"]
		self.city = serialized["city"]
		self.province = serialized["province"]
		self.organization = serialized["organization"]
		self.organization_unit = serialized["organization_unit"]
		self.country_code = serialized["country_code"]
		self.domain = serialized["domain"]
		#self.namecheap_username = serialized["namecheap_username"]
		#self.namecheap_api_key = serialized["namecheap_api_key"]
		#self.firebase_admin = serialized["firebase_admin"]
		#self.stripe_secret_key = serialized["stripe_secret_key"]
		#self.stripe_publishable_key = serialized["stripe_publishable_key"]
		#self.email_address = serialized["email_address"]
		#self.email_password = serialized["email_password"]
		self.email_smtp_host = serialized["email_smtp_host"]
		self.email_smtp_port = serialized["email_smtp_port"]
		self.purchase_tls_certificate = serialized["purchase_tls_certificate"]
		self.stripe_subscriptions = serialized["stripe_subscriptions"]
		self.stripe_products = serialized["stripe_products"]
		self.developers = serialized["developers"]
		self.remote = serialized['remote']
		self.vps_ip = serialized["vps_ip"]
		self.vps_port = serialized["vps_port"]
		self.vps_username = serialized["vps_username"]
		self._2fa = serialized["_2fa"]
		self.firebase_enabled = serialized["firebase_enabled"]
		self.stripe_enabled = serialized["stripe_enabled"]
		self.namecheap_enabled = serialized["namecheap_enabled"]
		self.email_enabled = serialized["email_enabled"]
		self.log_level = serialized["log_level"]
		self.users_subpath = serialized["users_subpath"]
		self.id_by_username = serialized["id_by_username"]
		self.styling = serialized["styling"]
		self.debug = serialized["debug"]

		# load secrets.
		local_security = security.Security(
			# the root path.
			root=self.root,)
		required = False
		migrations = dev0s.env.get("MIGRATIONS", format=bool, default=False)
		if self.firebase_enabled and not migrations: required = True
		self.firebase_admin = {
			"type":local_security.get_secret_env("FIREBASE_ADMIN_"+"type".upper(), default=None, required=required),
			"project_id":local_security.get_secret_env("FIREBASE_ADMIN_"+"project_id".upper(), default=None, required=required),
			"private_key_id":local_security.get_secret_env("FIREBASE_ADMIN_"+"private_key_id".upper(), default=None, required=required),
			"private_key":local_security.get_secret_env("FIREBASE_ADMIN_"+"private_key".upper(), default=None, required=required),
			"client_email":local_security.get_secret_env("FIREBASE_ADMIN_"+"client_email".upper(), default=None, required=required),
			"client_id":local_security.get_secret_env("FIREBASE_ADMIN_"+"client_id".upper(), default=None, required=required),
			"auth_uri":local_security.get_secret_env("FIREBASE_ADMIN_"+"auth_uri".upper(), default=None, required=required),
			"token_uri":local_security.get_secret_env("FIREBASE_ADMIN_"+"token_uri".upper(), default=None, required=required),
			"auth_provider_x509_cert_url":local_security.get_secret_env("FIREBASE_ADMIN_"+"auth_provider_x509_cert_url".upper(), default=None, required=required),
			"client_x509_cert_url":local_security.get_secret_env("FIREBASE_ADMIN_"+"client_x509_cert_url".upper(), default=None, required=required),
		}
		self.firebase_js = {
			"api_key":local_security.get_secret_env("FIREBASE_JS_"+"API_KEY", default=None, required=required),
			"auth_domain":local_security.get_secret_env("FIREBASE_JS_"+"AUTH_DOMAIN", default=None, required=required),
			"database_url":local_security.get_secret_env("FIREBASE_JS_"+"DATABASE_URL", default=None, required=False),
			"project_id":local_security.get_secret_env("FIREBASE_JS_"+"PROJECT_ID", default=None, required=required),
			"storage_bucket":local_security.get_secret_env("FIREBASE_JS_"+"STORAGE_BUCKET", default=None, required=required),
			"messaging_sender_id":local_security.get_secret_env("FIREBASE_JS_"+"MESSAGING_SENDER_ID", default=None, required=required),
			"app_id":local_security.get_secret_env("FIREBASE_JS_"+"APP_ID", default=None, required=required),
			"measurement_id":local_security.get_secret_env("FIREBASE_JS_"+"MEASUREMENT_ID", default=None, required=False),
		}
		self.stripe_secret_key = local_security.get_secret_env("STRIPE_SECRET_KEY", default=None, required=self.stripe_enabled)
		self.stripe_publishable_key = local_security.get_secret_env("STRIPE_PUBLISHABLE_KEY", default=None, required=self.stripe_enabled)
		self.email_address = local_security.get_secret_env("EMAIL_ADDRESS", default=None, required=self.email_enabled)
		self.email_password = local_security.get_secret_env("EMAIL_PASSWORD", default=None, required=self.email_enabled)
		self.aes_passphrase = local_security.get_secret_env("AES_PASSPHRASE", default=None, required=not migrations)
		self.namecheap_username = local_security.get_secret_env("NAMECHEAP_USERNAME", default=self.namecheap_enabled)
		self.namecheap_api_key = local_security.get_secret_env("NAMECHEAP_API_KEY", default=self.namecheap_enabled)

		# success.
		return True

		#

	# system functions.
	def __str__(self, indent=0):
		def __indent__(indent):
			s = ""
			for i in range(indent): s += " "
			return s
		serialized = self.serialize(save=False)
		_serialized_ = {}
		_serialized_["root"] = serialized["root"]
		_serialized_["name"] = serialized["name"]
		_serialized_["email"] = serialized["email"]
		_serialized_["country_code"] = serialized["country_code"]
		_serialized_["city"] = serialized["city"]
		_serialized_["organization"] = serialized["organization"]
		_serialized_["domain"] = serialized["domain"]
		_indent_ = __indent__(indent)
		str = json.dumps(_serialized_, indent=indent).replace('": "', ': ').replace('": ', ': ').replace('",', "").replace(f'\n{_indent_}"', f'\n{_indent_} * ').replace("{\n", "").replace('"\n}', "").replace("\n}", "").replace("{", "")
		return f"<w3bsite.Website; \n{str}\n>"
	
	#

#




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# classes imports.
from classes.config import *
from classes import utils
from apps.client import client
from apps.backups import backups

# the users control class.
class Manager(object):
	def __init__(self):

		# objects.
		self.client = client.client
		self.backups = backups.backups
	
		# service.	
		Files.chmod(f"{SOURCE}/__defaults__/start", permission="+x")
		self.service = dev0s.system.Service(
			id="vbackups",
			user=dev0s.defaults.vars.user,
			group=None,
			start=[f"{SOURCE}/__defaults__/start"],
			description="VBackups Web Interface",
			restart=True,
			restart_limit=5,
			restart_delay=10,
			logs=Files.join(DATABASE, "logs/logs"),
			errors=Files.join(DATABASE, "logs/errors"),
			traceback="dev0s.system.Service", )
		response = self.service.check()
		if not response.success: response.crash(json=dev0s.defaults.options.json)
		
		#

	# encryption.
	def install_encryption(self, passphrase=None, verify_passphrase=None, interactive=False):

		# generate.
		response = ssht00ls_agent.generate(
			passphrase=passphrase, 
			verify_passphrase=verify_passphrase,
			interactive=interactive,)
		if not response.success: return response

		# check client.
		response = self.client.initialize()
		if not response.success: return response
		
		# response.
		return dev0s.response.success("Successfully installed and activated the encryption.", log_level=0)

		#
	def activate_encryption(self, passphrase=None, interactive=False):
		
		# activate.
		response = ssht00ls_agent.activate(
			passphrase=passphrase, 
			interactive=interactive)
		if not response.success: return response

		# check client.
		response = self.client.initialize()
		if not response.success: return response

		# response.
		return dev0s.response.success("Successfully activated the encryption.", log_level=0)
		
		#

	# properties.
	@property
	def installed(self):
		return self.__installed__()
	def __installed__(self, filter=None, options=["django", "database", "client", "settings", "encryption"]):
		if filter == None:
			for i in options:
				b = self.__installed__(filter=i)
				if dev0s.defaults.options.log_level >= 3: print(f"@installed {i}: {b}")
				if not b:
					return False
			return True
		elif filter == "django":
			return website.users.exists(username=ADMINISTRATOR, filter="django")
		elif filter == "database":
			return Files.exists(DATABASE.join("data/db.sqlite3"))
		elif filter == "client":
			return Files.exists(DATABASE.join(f"clients/{ADMINISTRATOR}/private_key")) and Files.exists(DATABASE.join(f"clients/{ADMINISTRATOR}/public_key")) and self.client.client.alias in ssht00ls.clients
		elif filter == "settings":
			return None not in [ self.client.settings["server"]["name"], self.client.settings["server"]["domain"] ]
		elif filter == "encryption":
			return ssht00ls_agent.generated
		else: raise ValueError(f"Invalid usage, filter [{filter}] is not a valid option, options: {Array(options).string(joiner=' ')}.")
	
	#

# initialized classes.
manager = Manager()

